import plotly
import plotly.express as px
from plotly.graph_objects import Figure as _Figure
from pandas import DataFrame


def render_3D_scatter_plotly(embeddings_dataframe: DataFrame) -> _Figure:
    """
    Return a Plotly Figure (3D scatter plot) based on a DataFrame containing three components.

    :param embeddings_dataframe: the DataFrame *must* contain three numerical columns called `component_0`,
            `component_1` and `component_2`. The DataFrame index will be used to identify the points in the
            scatter plot. Optionally, the DataFrame may contain a column called `label` which will be used
            to color the points in the scatter plot.
    :return: A 3D scatter plot
    """

    if 'label' in embeddings_dataframe.columns:
        fig = px.scatter_3d(embeddings_dataframe,
                            x='component_0',
                            y='component_1',
                            z='component_2',
                            color='label',
                            symbol='label',
                            hover_name=embeddings_dataframe.index,
                            hover_data=["label"]
                            )
    else:
        fig = px.scatter_3d(embeddings_dataframe,
                            x='component_0',
                            y='component_1',
                            z='component_2',
                            hover_name=embeddings_dataframe.index,
                            )

    fig.update_layout(
        # Remove axes ticks and labels as they are usually not informative
        scene=dict(
            xaxis=dict(
                showticklabels=False,
                showspikes=False,
                title=""
            ),
            yaxis=dict(
                showticklabels=False,
                showspikes=False,
                title=""
            ),
            zaxis=dict(
                showticklabels=False,
                showspikes=False,
                title=""
            )
        ),
    )

    return fig


def render_scatter_plotly(embeddings_dataframe: DataFrame) -> _Figure:
    """
    Return a Plotly Figure (2D scatter plot) based on a DataFrame containing three components.

    :param embeddings_dataframe: the DataFrame *must* contain two numerical columns called `component_0`
            and `component_1`. The DataFrame index will be used to identify the points in the
            scatter plot. Optionally, the DataFrame may contain a column called `label` which will be used
            to color the points in the scatter plot.
    :return: A 2D scatter plot
    """

    if 'label' in embeddings_dataframe.columns:
        fig = px.scatter(embeddings_dataframe,
                         x='component_0',
                         y='component_1',
                         color='label',
                         symbol='label',
                         hover_name=embeddings_dataframe.index,
                         hover_data=["label"]
                         )
    else:
        fig = px.scatter(embeddings_dataframe,
                         x='component_0',
                         y='component_1',
                         hover_name=embeddings_dataframe.index,
                         )

    fig.update_layout(
        # Remove axes ticks and labels as they are usually not informative
        scene=dict(
            xaxis=dict(
                showticklabels=False,
                showspikes=False,
                title=""
            ),
            yaxis=dict(
                showticklabels=False,
                showspikes=False,
                title=""
            )
        ),
    )

    return fig


def save_plotly_figure_to_html(figure: _Figure, path: str) -> None:
    """
    Store plotly figure as interactive HTML file

    :param figure: A Plotly Figure
    :param path: A string representing the path and/or filename where the HTML figure should be stored
         (e.g.: /path/to/figure.html).
    """
    plotly.offline.plot(figure, filename=path)



from enum import Enum


class Disorder(Enum):
    DISORDER = 'X'
    ORDER = '-'
    UNKNOWN = '?'

    def __str__(self):
        return {
            self.DISORDER: "Disorder",
            self.ORDER: "Order",
            self.UNKNOWN: "Unknown"
        }.get(self)

    @staticmethod
    def isAAFeature():
        return True



import torch.nn as nn


class SUBCELL_FNN(nn.Module):
    # in 10 states and a binary classification into membrane-bound vs. soluble
    def __init__(self, use_batch_norm=True):
        super(SUBCELL_FNN, self).__init__()
        # Linear layer, taking embedding dimension 1024 to make predictions:
        if use_batch_norm:
            self.layer = nn.Sequential(
                nn.Linear(1024, 32),  # in, out
                nn.Dropout(0.25),  # dropout
                nn.ReLU(),
                nn.BatchNorm1d(32)
                )
        else:
            self.layer = nn.Sequential(
                nn.Linear(1024, 32),  # in, out
                nn.Dropout(0.25),  # dropout
                nn.ReLU(),
                )
        self.loc_classifier = nn.Linear(32, 10)
        self.mem_classifier = nn.Linear(32, 2)

    def forward(self, x):
        out = self.layer(x)  # map 1024-dimensional ELMo vector to 32-dims
        # based on 32 dims, predict localization and membrane-bound
        Yhat_loc = self.loc_classifier(out)
        Yhat_mem = self.mem_classifier(out)

        return Yhat_loc, Yhat_mem


class SECSTRUCT_CNN(nn.Module):
    # Convolutional neural network for prediction of Sec.Struct. in 3- & 8-states and disorder
    def __init__(self):
        super(SECSTRUCT_CNN, self).__init__()

        self.elmo_feature_extractor = nn.Sequential(
            nn.Conv2d(1024, 32, kernel_size=(7, 1), padding=(3, 0)),
            nn.ReLU(),
            nn.Dropout(0.25),
        )

        self.dssp3_classifier = nn.Sequential(
            nn.Conv2d(32, 3, kernel_size=(7, 1), padding=(3, 0))
        )
        self.dssp8_classifier = nn.Sequential(
            nn.Conv2d(32, 8, kernel_size=(7, 1), padding=(3, 0))
        )
        self.diso_classifier = nn.Sequential(
            nn.Conv2d(32, 2, kernel_size=(7, 1), padding=(3, 0))
        )

    def forward(self, x):
        x = self.elmo_feature_extractor(x)  # compress ELMo features to 32-dims

        d3_Yhat = self.dssp3_classifier(x)
        d8_Yhat = self.dssp8_classifier(x)
        diso_Yhat = self.diso_classifier(x)

        return d3_Yhat, d8_Yhat, diso_Yhat



import logging
import os
from os import path as os_path
from pathlib import Path

from bio_embeddings.utilities.filemanagers.FileManagerInterface import FileManagerInterface

logger = logging.getLogger(__name__)


class FileSystemFileManager(FileManagerInterface):

    def __init__(self):
        super().__init__()

    def exists(self, prefix, stage=None, file_name=None, extension=None) -> bool:
        path = Path(prefix)

        if stage:
            path /= stage
        if file_name:
            path /= file_name + (extension or "")

        return os_path.exists(path)

    def get_file(self, prefix, stage, file_name, extension=None) -> str:
        path = Path(prefix)

        if stage:
            path /= stage
        if file_name:
            path /= file_name + (extension or "")

        return str(path)

    def create_file(self, prefix, stage, file_name, extension=None) -> str:
        path = Path(prefix)

        if stage:
            path /= stage

        path /= file_name + (extension or "")

        try:
            with open(path, 'w'):
                os.utime(path, None)
        except OSError as e:
            logger.error("Failed to create file %s" % path)
            raise e
        else:
            logger.info("Created the file %s" % path)

        return str(path)

    def create_directory(self, prefix, stage, directory_name) -> str:
        path = Path(prefix)

        if stage:
            path /= stage

        path /= directory_name

        try:
            os.mkdir(path)
        except FileExistsError:
            logger.info("Directory %s already exists." % path)
        except OSError as e:
            logger.error("Failed to create directory %s" % path)
            raise e
        else:
            logger.info("Created the directory %s" % path)

        return str(path)

    def create_stage(self, prefix, stage) -> str:
        path = Path(prefix) / stage

        try:
            os.mkdir(path)
        except FileExistsError:
            logger.info("Stage directory %s already exists." % path)
        except OSError as e:
            logger.error("Failed to create stage directory %s" % path)
            raise e
        else:
            logger.info("Created the stage directory %s" % path)

        return str(path)

    def create_prefix(self, prefix) -> str:
        path = Path(prefix)

        try:
            os.mkdir(path)
        except FileExistsError:
            logger.info("Prefix directory %s already exists." % path)
        except OSError as e:
            logger.error("Failed to create prefix directory %s" % path)
            raise e
        else:
            logger.info("Created the prefix directory %s" % path)

        return str(path)



import logging
import shutil
import h5py
import numpy

from copy import deepcopy
from typing import Dict, Any

from Bio import SeqIO
from pandas import read_csv, DataFrame
from tqdm import tqdm

from bio_embeddings.embed import name_to_embedder, EmbedderInterface
from bio_embeddings.utilities import (
    FileManagerInterface,
    InvalidParameterError,
    check_required,
    get_file_manager,
    get_model_directories_from_zip,
    get_model_file,
)
from bio_embeddings.utilities.backports import nullcontext

logger = logging.getLogger(__name__)


def _print_expected_file_sizes(
    embedder: EmbedderInterface, mapping_file: DataFrame, result_kwargs: Dict[str, Any]
) -> None:
    """
    Logs the lower bound size of embeddings_file and reduced_embedding_file

    :param embedder: the embedder being used
    :param mapping_file: the mapping file of the sequences
    :param result_kwargs: the kwargs passed to the pipeline --> will decide what to print

    :return: Nothing.
    """
    per_amino_acid_size_in_bytes = 4 * embedder.embedding_dimension * embedder.number_of_layers
    per_protein_size_in_bytes = 4 * embedder.embedding_dimension

    total_number_of_proteins = len(mapping_file)
    total_aa = mapping_file['sequence_length'].sum()

    embeddings_file_size_in_MB = per_amino_acid_size_in_bytes * total_aa * pow(10, -6)
    reduced_embeddings_file_size_in_MB = per_protein_size_in_bytes * total_number_of_proteins * pow(10, -6)

    required_space_in_MB = 0

    if result_kwargs.get("reduce") is True:
        logger.info(f"The minimum expected size for the reduced_embedding_file is "
                    f"{reduced_embeddings_file_size_in_MB:.3f}MB.")

        required_space_in_MB += reduced_embeddings_file_size_in_MB

    # TODO: calculate size of transformed embeddings?
    if not (result_kwargs.get("reduce") is True and result_kwargs.get("discard_per_amino_acid_embeddings") is True):
        logger.info(f"The minimum expected size for the embedding_file is {embeddings_file_size_in_MB:.3f}MB.")

        required_space_in_MB += embeddings_file_size_in_MB

    _, _, available_space_in_bytes = shutil.disk_usage(result_kwargs.get('prefix'))

    available_space_in_MB = available_space_in_bytes * pow(10, -6)

    if available_space_in_MB < required_space_in_MB:
        logger.warning(f"You are attempting to generate {required_space_in_MB:.3f}MB worth of embeddings, "
                       f"but only {available_space_in_MB:.3f}MB are available at "
                       f"the prefix({result_kwargs.get('prefix')}). \n"
                       f"We suggest you stop execution NOW and double check you have enough free space available. "
                       f"Alternatively, try reducing the input FASTA file.")
    else:
        logger.info(f"You are going to generate a total of {required_space_in_MB:.3f}MB of embeddings, and have "
                    f"{available_space_in_MB:.3f}MB available at {result_kwargs.get('prefix')}.")


def _get_reduced_embeddings_file_context(
    file_manager: FileManagerInterface, result_kwargs: Dict[str, Any]
):
    """
    :param file_manager: The FileManager derived class which will be used to create the file
    :param result_kwargs: A dictionary which will be updated in-place to include the path to the newly created file

    :return: a file context
    """

    # Create reduced embeddings file if set in params
    result_kwargs.setdefault("reduce", False)

    if result_kwargs["reduce"] is True:
        reduced_embeddings_file_path = file_manager.create_file(
            result_kwargs.get("prefix"),
            result_kwargs.get("stage_name"),
            "reduced_embeddings_file",
            extension=".h5",
        )
        result_kwargs["reduced_embeddings_file"] = reduced_embeddings_file_path
        return h5py.File(reduced_embeddings_file_path, "w")

    return nullcontext()


def _get_embeddings_file_context(
    file_manager: FileManagerInterface, result_kwargs: Dict[str, Any]
):
    """
    :param file_manager: The FileManager derived class which will be used to create the file
    :param result_kwargs: A dictionary which will be updated in-place to include the path to the newly created file

    :return: a file context
    """

    result_kwargs.setdefault("discard_per_amino_acid_embeddings", False)

    if result_kwargs["discard_per_amino_acid_embeddings"] is True:
        if result_kwargs.get("reduce", False) is False and result_kwargs.get("embeddings_transformer_function") is None:
            raise InvalidParameterError(
                "Cannot only have discard_per_amino_acid_embeddings: True. "
                "Either also set `reduce: True` or define an `embeddings_transformer_function`, or both."
            )
        return nullcontext()
    else:
        embeddings_file_path = file_manager.create_file(
            result_kwargs.get("prefix"),
            result_kwargs.get("stage_name"),
            "embeddings_file",
            extension=".h5",
        )
        result_kwargs["embeddings_file"] = embeddings_file_path
        return h5py.File(embeddings_file_path, "w")


def _get_transformed_embeddings_file_context(
    file_manager: FileManagerInterface, result_kwargs: Dict[str, Any]
):
    """
    :param file_manager: The FileManager derived class which will be used to create the file
    :param result_kwargs: A dictionary which will be updated in-place to include the path to the newly created file

    :return: a file context
    """

    result_kwargs.setdefault("embeddings_transformer_function", None)

    if result_kwargs["embeddings_transformer_function"] is not None:
        transformed_embeddings_file_path = file_manager.create_file(
            result_kwargs.get("prefix"),
            result_kwargs.get("stage_name"),
            "transformed_embeddings_file",
            extension=".h5",
        )
        result_kwargs["transformed_embeddings_file"] = transformed_embeddings_file_path
        return h5py.File(transformed_embeddings_file_path, "w")

    return nullcontext()


def _check_transform_embeddings_function(embedder: EmbedderInterface, result_kwargs: Dict[str, Any]):

    result_kwargs.setdefault("embeddings_transformer_function", None)

    if result_kwargs["embeddings_transformer_function"] is not None:
        try:
            transform_function = eval(result_kwargs["embeddings_transformer_function"], {}, {"np": numpy})
        except TypeError:
            raise InvalidParameterError(f"`embeddings_transformer_function` must be callable! \n"
                                        f"Instead is {result_kwargs['embeddings_transformer_function']}\n"
                                        f"Most likely you want a lambda function.")

        if not callable(transform_function):
            raise InvalidParameterError(f"`embeddings_transformer_function` must be callable! \n"
                                        f"Instead is {result_kwargs['embeddings_transformer_function']}\n"
                                        f"Most likely you want a lambda function.")

        template_embedding = embedder.embed("SEQVENCE")

        # Check that it works in principle
        try:
            transformed_template_embedding = transform_function(template_embedding)
        except:
            raise InvalidParameterError(f"`embeddings_transformer_function` must be valid callable! \n"
                                        f"Instead is {result_kwargs['embeddings_transformer_function']}\n"
                                        f"This function excepts when processing an embedding.")

        # Check that return can be cast to np.array
        try:
            numpy.array(transformed_template_embedding)
        except:
            raise InvalidParameterError(f"`embeddings_transformer_function` must be valid callable "
                                        f"returning numpy array compatible object! \n"
                                        f"Instead is {result_kwargs['embeddings_transformer_function']}\n"
                                        f"This function excepts when processing an embedding.")


def embed_and_write_batched(
    embedder: EmbedderInterface,
    file_manager: FileManagerInterface,
    result_kwargs: Dict[str, Any],
    half_precision: bool = False
) -> Dict[str, Any]:
    """ The shared code between the SeqVec, Albert, Bert and XLNet pipelines """
    # Lazy fasta file reader. The mapping file contains the corresponding ids in the same order
    sequences = (
        str(entry.seq)
        for entry in SeqIO.parse(result_kwargs["remapped_sequences_file"], "fasta")
    )
    # We want to read the unnamed column 0 as str (esp. with simple_remapping), which requires some workarounds
    # https://stackoverflow.com/a/29793294/3549270
    mapping_file = read_csv(result_kwargs["mapping_file"], index_col=0)
    mapping_file.index = mapping_file.index.astype('str')

    # Print the minimum required file sizes
    _print_expected_file_sizes(embedder, mapping_file, result_kwargs)

    # Get transformer function, if available
    transform_function = result_kwargs.get("embeddings_transformer_function", None)

    if transform_function:
        transform_function = eval(transform_function, {}, {"np": numpy})

    # Open embedding files or null contexts and iteratively save embeddings to file
    with _get_embeddings_file_context(
        file_manager, result_kwargs
    ) as embeddings_file, _get_reduced_embeddings_file_context(
        file_manager, result_kwargs
    ) as reduced_embeddings_file, _get_transformed_embeddings_file_context(
        file_manager, result_kwargs
    ) as transformed_embeddings_file:
        embedding_generator = embedder.embed_many(
            sequences, result_kwargs.get("max_amino_acids")
        )
        for sequence_id, original_id, embedding in zip(
            mapping_file.index,
            mapping_file["original_id"],
            tqdm(embedding_generator, total=len(mapping_file))
        ):
            # embedding: numpy.ndarray
            if half_precision:
                embedding = embedding.astype(numpy.float16)
            if result_kwargs.get("discard_per_amino_acid_embeddings") is False:
                dataset = embeddings_file.create_dataset(sequence_id, data=embedding)
                dataset.attrs["original_id"] = original_id
            if result_kwargs.get("reduce") is True:
                dataset = reduced_embeddings_file.create_dataset(
                    sequence_id, data=embedder.reduce_per_protein(embedding)
                )
                dataset.attrs["original_id"] = original_id
            if transform_function:
                dataset = transformed_embeddings_file.create_dataset(
                    sequence_id, data=numpy.array(transform_function(embedding))
                )
                dataset.attrs["original_id"] = original_id

    return result_kwargs


# Some of this might not be available when installed without the `all` extra
ALL_PROTOCOLS = [
    "bepler",
    "cpcprot",
    "esm",
    "plus_rnn",
    "prottrans_albert_bfd",
    "prottrans_bert_bfd",
    "prottrans_t5_bfd",
    "prottrans_xlnet_uniref100",
    "seqvec",
    "unirep",
]

# TODO: 10000 is a random guess
# There remainder was measured for a GTX 1080 with 8GB memory
DEFAULT_MAX_AMINO_ACIDS = {
    "bepler": 10000,
    "cpcprot": 10000,
    "esm": 10000,
    "plus_rnn": 10000,
    "prottrans_albert_bfd": 3035,
    "prottrans_bert_bfd": 6024,
    "prottrans_t5_bfd": 1000,
    "prottrans_xlnet_uniref100": 4000,
    "seqvec": 15000,
    "unirep": 10000,
}


def run(**kwargs):
    """
    Run embedding protocol

    Parameters
    ----------
    kwargs arguments (* denotes optional):
        sequences_file: Where sequences live
        prefix: Output prefix for all generated files
        protocol: Which embedder to use
        mapping_file: the mapping file generated by the pipeline when remapping indexes
        stage_name: The stage name

    Returns
    -------
    Dictionary with results of stage
    """
    check_required(
        kwargs,
        ["protocol", "prefix", "stage_name", "remapped_sequences_file", "mapping_file"],
    )

    if kwargs["protocol"] not in name_to_embedder:
        if kwargs["protocol"] in ALL_PROTOCOLS:
            raise InvalidParameterError(
                f"The extra for the protocol {kwargs['protocol']} is missing. "
                "See https://docs.bioembeddings.com/#installation on how to install all extras"
            )
        raise InvalidParameterError(
            "Invalid protocol selection: {}. Valid protocols are: {}".format(
                kwargs["protocol"], ", ".join(name_to_embedder.keys())
            )
        )

    embedder_class = name_to_embedder[kwargs["protocol"]]

    if kwargs["protocol"] == "unirep" and kwargs.get("use_cpu") is not None:
        raise InvalidParameterError("UniRep does not support configuring `use_cpu`")

    result_kwargs = deepcopy(kwargs)

    # Download necessary files if needed
    # noinspection PyProtectedMember
    for file in embedder_class._necessary_files:
        if not result_kwargs.get(file):
            result_kwargs[file] = get_model_file(model=embedder_class.name, file=file)

    # noinspection PyProtectedMember
    for directory in embedder_class._necessary_directories:
        if not result_kwargs.get(directory):
            result_kwargs[directory] = get_model_directories_from_zip(
                model=embedder_class.name, directory=directory
            )

    result_kwargs.setdefault("max_amino_acids", DEFAULT_MAX_AMINO_ACIDS[kwargs["protocol"]])

    file_manager = get_file_manager(**kwargs)
    embedder: EmbedderInterface = embedder_class(**result_kwargs)
    _check_transform_embeddings_function(embedder, result_kwargs)

    return embed_and_write_batched(embedder, file_manager, result_kwargs, kwargs.get("half_precision", False))



import numpy as np
import h5py
import logging
import torch

from enum import Enum
from hashlib import md5
from typing import List, Union
from pandas import DataFrame, read_csv

from Bio import SeqIO
from Bio.SeqRecord import SeqRecord

from bio_embeddings.utilities.exceptions import MissingParameterError, ConversionUniqueMismatch


logger = logging.getLogger(__name__)


def get_device(device: Union[None, str, torch.device] = None) -> torch.device:
    """Returns what the user specified, or defaults to the GPU,
    with a fallback to CPU if no GPU is available."""
    if isinstance(device, torch.device):
        return device
    elif device:
        return torch.device(device)
    elif torch.cuda.is_available():
        return torch.device("cuda")
    else:
        return torch.device("cpu")


def check_required(params: dict, keys: List[str]):
    """
    Verify if required set of parameters is present in configuration

    Parameters
    ----------
    params : dict
        Dictionary with parameters
    keys : list-like
        Set of parameters that has to be present in params

    Raises
    ------
    MissingParameterError
    """
    missing = [k for k in keys if k not in params]

    if len(missing) > 0:
        raise MissingParameterError(
            "Missing required parameters: {} \nGiven: {}".format(
                ", ".join(missing), params
            )
        )


def _assign_hash(sequence_record: SeqRecord) -> SeqRecord:
    sequence_record.id = md5(str(sequence_record.seq).encode()).hexdigest()

    return sequence_record


def read_fasta(path: str) -> List[SeqRecord]:
    """
    Helper function to read FASTA file.

    :param path: path to a valid FASTA file
    :return: a list of SeqRecord objects.
    """
    try:
        return list(SeqIO.parse(path, "fasta"))
    except FileNotFoundError:
        raise  # Already says "No such file or directory"
    except Exception as e:
        raise ValueError(f"Could not parse '{path}'. Are you sure this is a valid fasta file?") from e


def reindex_sequences(
        sequence_records: List[SeqRecord], simple=False
) -> (SeqRecord, DataFrame):
    """
    Function will sort and re-index the sequence_records IN PLACE! (change the original list!).
    Returns a DataFrame with the mapping.

    :param sequence_records: List of sequence records
    :param simple: Boolean; if set to true use numerical index (1,2,3,4) instead of md5 hash
    :return: A dataframe with the mapping with key the new ids and a column "original_id" containing the previous id, and the sequence length.
    """
    sequence_records[:] = sorted(sequence_records, key=lambda seq: -len(seq))
    original_ids = [s.id for s in sequence_records]

    if simple:
        new_ids = list()
        for id, record in enumerate(sequence_records):
            record.id = str(id)
            new_ids.append(str(id))
    else:
        sequence_records[:] = map(_assign_hash, sequence_records)
        new_ids = [s.id for s in sequence_records]

    df = DataFrame(
        zip(original_ids, [len(seq) for seq in sequence_records]),
        columns=["original_id", "sequence_length"],
        index=new_ids,
    )

    return df


def write_fasta_file(sequence_records: List[SeqRecord], file_path: str) -> None:
    SeqIO.write(sequence_records, file_path, "fasta")


def convert_list_of_enum_to_string(list_of_enums: List[Enum]) -> str:
    return "".join([e.value for e in list_of_enums])


def reindex_h5_file(h5_file_path: str, mapping_file_path: str):
    """
    Will rename the dataset keys using the "original_id" from the mapping file.
    This operation is generally considered unsafe, as the "original_id" is unsafe
    (may contain invalid characters, duplicates, or empty strings).

    Some sanity checks are performed before starting the renaming process,
    but generally applying this function is discouraged unless you know what you are doing.

    :param h5_file_path: path to the hd5_file to re-index
    :param mapping_file_path: path to the mapping file (this must have the first column be the current keys, and a column "original_id" as the new desired id)
    :return: Nothing -- conversion happens in place!
    """

    mapping_file = read_csv(mapping_file_path, index_col=0)
    mapping_file.index = mapping_file.index.map(str)
    mapping_file['original_id'] = mapping_file['original_id'].astype(str)
    conversion_table = list(zip(mapping_file.index.values, mapping_file['original_id'].values))
    unique_froms = set([e[0] for e in conversion_table])
    unique_tos = set([e[1] for e in conversion_table if e])

    if len(unique_froms) != len(unique_tos):
        raise ConversionUniqueMismatch(f"Conversion unique count mismatch.\n"
                                       f"Your mapping file contains {len(unique_froms)} unique ids, which you are truing to convert to {len(unique_tos)} unique original_ids.\n"
                                       f"These numbers *must* match. You likely have: duplicate original_id's, or empty strings in original_id.")

    with h5py.File(h5_file_path, "r+") as h5_file:
        keys_set = set(h5_file.keys())
        unchanged_set = keys_set - unique_froms

        if len(unchanged_set) > 0:
            logger.warning(f"There are some keys in your h5 file which won't be re-indexed!\n"
                           f"These are: {unchanged_set}.")

        changeable_set = unique_froms.union(keys_set)

        if len(changeable_set) == 0:
            logger.info("Nothing was re-indexed.")

        else:
            logger.info(f"Reindexing the following keys: {changeable_set}")

            for (from_index, to_index) in filter(lambda item: item[0] in keys_set, conversion_table):
                h5_file.move(from_index, to_index)


def remove_identifiers_from_annotations_file(faulty_identifiers: list, annotation_file_path: str) -> DataFrame:
    """
    Removes id
    :param faulty_identifiers: a list of identifiers
    :param annotation_file_path: a str detailing the path
    :return: a new DataFrame with the annotations removed
    """

    annotation_file = read_csv(annotation_file_path)
    return annotation_file[annotation_file['identifier'].isin(
        set(annotation_file['identifier'].values) - set(faulty_identifiers)
    )]


class QueryEmbeddingsFile:
    """
    A helper class that allows you to retrieve embeddings from an embeddings file based on either the `original_id`
    (extracted from the FASTA header during the embed stage), or via the `new_id` (assigned during the embed stage,
    either an MD5 hash of the input sequence, or an integer (if `remapping_simple: True`).

    Available for embeddings created with the pipeline starting with v0.1.5
    """

    def __init__(self, embeddings_file: h5py.File):
        """
        :param embeddings_file: an h5py File, aka `h5py.File("/path/to/file.h5")`.
        """
        self._lookup_table = dict(
            (embeddings_file[new_id].attrs["original_id"], new_id) for new_id in embeddings_file.keys()
        )

        self._embeddings_file = embeddings_file

    def query_original_id(self, original_id: str) -> np.array:
        """
        Query embeddings file using the original id, aka. the string extracted from the FASTA header of the sequence.

        :param original_id: a string representing the id extracted from the FASTA header
        :return: the embedding as a numpy array
        """
        return np.array(self._embeddings_file[self._lookup_table[original_id]])

    def query_new_id(self, new_id: str) -> np.array:
        """
        Query embeddings file using the new id, aka. either the MD5 hash of the sequence or a number.

        :param new_id: a string representing the new id.
        :return: the embedding as a numpy array
        """
        return np.array(self._embeddings_file[new_id])



import logging
import h5py
import numpy as np

from copy import deepcopy
from typing import Dict, Any, List
from Bio.Seq import Seq
from pandas import read_csv, DataFrame, concat as concatenate_dataframe
from sklearn.metrics import pairwise_distances as _pairwise_distances

from bio_embeddings.extract.basic import BasicAnnotationExtractor
from bio_embeddings.extract.unsupervised_utilities import get_k_nearest_neighbours
from bio_embeddings.utilities.remote_file_retriever import get_model_file
from bio_embeddings.utilities.filemanagers import get_file_manager
from bio_embeddings.utilities.helpers import check_required, read_fasta, convert_list_of_enum_to_string, \
    write_fasta_file
from bio_embeddings.utilities.exceptions import InvalidParameterError, UnrecognizedEmbeddingError, \
    InvalidAnnotationFileError

logger = logging.getLogger(__name__)


def _flatten_2d_list(l: List[List[str]]) -> List[str]:
    return [item for sublist in l for item in sublist]


def unsupervised(**kwargs) -> Dict[str, Any]:
    check_required(kwargs, ['reference_embeddings_file', 'reference_annotations_file', 'reduced_embeddings_file'])

    result_kwargs = deepcopy(kwargs)
    file_manager = get_file_manager(**kwargs)

    # Try to create final files (if this fails, now is better than later
    transferred_annotations_file_path = file_manager.create_file(result_kwargs.get('prefix'),
                                                                 result_kwargs.get('stage_name'),
                                                                 'transferred_annotations_file',
                                                                 extension='.csv')

    # Read the reference annotations and reference embeddings

    # The reference annotations file must be CSV containing two columns & headers like:
    # identifier,label
    # ** identifier doesn't need to be unique **
    reference_annotations_file = read_csv(result_kwargs['reference_annotations_file'])

    # If reference annotations contain nans (either in label or identifier) throw an error!
    # https://github.com/sacdallago/bio_embeddings/issues/58
    # https://datatofish.com/check-nan-pandas-dataframe/
    if reference_annotations_file[['identifier', 'label']].isnull().values.any():
        raise InvalidAnnotationFileError("Your annotation file contains NaN values in either identifier or label columns.\n"
                                         "Please remove these and run the pipeline again.")

    # Save a copy of the annotation file with only necessary cols cols
    input_reference_annotations_file_path = file_manager.create_file(result_kwargs.get('prefix'),
                                                                     result_kwargs.get('stage_name'),
                                                                     'input_reference_annotations_file',
                                                                     extension='.csv')

    reference_annotations_file.to_csv(input_reference_annotations_file_path, index=False)

    result_kwargs['input_reference_annotations_file'] = input_reference_annotations_file_path

    # Starting from here order is super important!
    reference_identifiers = reference_annotations_file['identifier'].unique()
    reference_identifiers.sort()
    reference_embeddings = list()

    # Save a copy of the reference embeddings file with only necessary embeddings
    input_reference_embeddings_file_path = file_manager.create_file(result_kwargs.get('prefix'),
                                                                    result_kwargs.get('stage_name'),
                                                                    'input_reference_embeddings_file',
                                                                    extension='.h5')

    result_kwargs['input_reference_embeddings_file'] = input_reference_embeddings_file_path

    # Only read in embeddings for annotated sequences! This will save RAM/GPU_RAM.
    with h5py.File(result_kwargs['reference_embeddings_file'], 'r') as reference_embeddings_file:
        # Sanity check: check that all identifiers in reference_annotation_file are present as embeddings

        unembedded_identifiers = set(reference_identifiers) - set(reference_embeddings_file.keys())

        if len(unembedded_identifiers) > 0:
            raise UnrecognizedEmbeddingError("Your reference_annotations_file includes identifiers for which "
                                             "no embedding can be found in your reference_embeddings_file.\n"
                                             "We require the set of identifiers in the reference_annotations_file "
                                             "to be a equal or a subset of the embeddings present in the "
                                             "reference_embeddings_file.\n"
                                             "To fix this issue, you can use the "
                                             "bio_embeddings.utilities.remove_identifiers_from_annotations_file "
                                             "function (see notebooks). "
                                             "The faulty identifiers are:\n['" +
                                             "','".join(unembedded_identifiers) + "']")

        with h5py.File(result_kwargs['input_reference_embeddings_file'], 'w') as input_reference_embeddings_file:
            for identifier in reference_identifiers:
                current_embedding = np.array(reference_embeddings_file[identifier])
                reference_embeddings.append(current_embedding)
                input_reference_embeddings_file.create_dataset(identifier, data=current_embedding)

    # mapping file will be needed to transfer annotations
    mapping_file = read_csv(result_kwargs['mapping_file'], index_col=0)
    mapping_file.index = mapping_file.index.map(str)

    # Important to have consistent ordering!
    target_identifiers = mapping_file.index.values
    target_identifiers.sort()
    target_embeddings = list()

    with h5py.File(result_kwargs['reduced_embeddings_file'], 'r') as reduced_embeddings_file:
        for identifier in target_identifiers:
            target_embeddings.append(np.array(reduced_embeddings_file[identifier]))

    result_kwargs['n_jobs'] = result_kwargs.get('n_jobs', 1)
    result_kwargs['metric'] = result_kwargs.get('metric', 'euclidean')

    pairwise_distances = _pairwise_distances(
        target_embeddings,
        reference_embeddings,
        metric=result_kwargs['metric'],
        n_jobs=result_kwargs['n_jobs']
    )

    result_kwargs['keep_pairwise_distances_matrix_file'] = result_kwargs.get('keep_pairwise_distances_matrix_file', False)

    if result_kwargs['keep_pairwise_distances_matrix_file']:
        pairwise_distances_matrix_file_path = file_manager.create_file(result_kwargs.get('prefix'),
                                                                       result_kwargs.get('stage_name'),
                                                                       'pairwise_distances_matrix_file',
                                                                       extension='.csv')
        pairwise_distances_matrix_file = DataFrame(pairwise_distances,
                                                   index=target_identifiers,
                                                   columns=reference_identifiers)
        pairwise_distances_matrix_file.to_csv(pairwise_distances_matrix_file_path, index=True)
        result_kwargs['pairwise_distances_matrix_file'] = pairwise_distances_matrix_file_path

    # transfer & store annotations
    result_kwargs['k_nearest_neighbours'] = result_kwargs.get('k_nearest_neighbours', 1)

    k_nn_indices, k_nn_distances = get_k_nearest_neighbours(pairwise_distances, result_kwargs['k_nearest_neighbours'])

    k_nn_identifiers = list(map(reference_identifiers.__getitem__, k_nn_indices))
    k_nn_annotations = list()

    for row in k_nn_identifiers:
        k_nn_annotations.append([";".join(
                reference_annotations_file[reference_annotations_file['identifier'] == identifier]['label'].values
            ) for identifier in row])

    # At this stage I have: nxk list of identifiers (strings), nxk indices (ints), nxk distances (floats),
    # nxk annotations
    # Now I need to expand the lists into a table and store the table into a CSV

    k_nn_identifiers_df = DataFrame(k_nn_identifiers, columns=[f"k_nn_{i+1}_identifier" for i in range(len(k_nn_identifiers[0]))])
    k_nn_distances_df = DataFrame(k_nn_distances, columns=[f"k_nn_{i+1}_distance" for i in range(len(k_nn_distances[0]))])
    k_nn_annotations_df = DataFrame(k_nn_annotations, columns=[f"k_nn_{i+1}_annotations" for i in range(len(k_nn_annotations[0]))])

    transferred_annotations_dataframe = concatenate_dataframe([k_nn_identifiers_df, k_nn_distances_df, k_nn_annotations_df], axis=1)
    transferred_annotations_dataframe.index = target_identifiers

    # At this stage we would like to aggregate all k_nn_XX_annotations into one column
    # -  A row in the k_nn_annotations matrix is string with X annotations (e.g. ["A;B", "A;C", "D"])
    # -  Each annotation in the string is separated by a ";"
    # Thus:
    # 1. Join all strings in a row separating them with ";" (aka ["A;B", "C"] --> "A;B;A;C;D")
    # 2. Split joined string into separate annotations using split(";") (aka "A;B;A;C;D" --> ["A","B","A","C","D"])
    # 3. Take a unique set of annotations by using set(*) (aka ["A","B","A","C","D"] --> set{"A","B","C","D"})
    # 4. Join the new unique set of annotations using ";" (aka set{"A","B","C","D"}) --> "A;B;C;D")
    transferred_annotations_dataframe['transferred_annotations'] = [";".join(set(";".join(k_nn_row).split(";"))) for k_nn_row in k_nn_annotations]

    # Merge with mapping file! Get also original ids!
    transferred_annotations_dataframe = mapping_file.join(transferred_annotations_dataframe)
    transferred_annotations_dataframe.to_csv(transferred_annotations_file_path, index=True)

    result_kwargs['transferred_annotations_file'] = transferred_annotations_file_path

    return result_kwargs


def seqvec_from_publication(**kwargs) -> Dict[str, Any]:
    return predict_annotations_using_basic_models("seqvec_from_publication", **kwargs)


def bert_from_publication(**kwargs) -> Dict[str, Any]:
    return predict_annotations_using_basic_models("bert_from_publication", **kwargs)


def predict_annotations_using_basic_models(model, **kwargs) -> Dict[str, Any]:
    """
    Protocol extracts secondary structure (DSSP3 and DSSP8), disorder, subcellular location and membrane boundness
    from "embeddings_file". Embeddings can either be generated with SeqVec or ProtBert.
    SeqVec models are used in this publication: https://doi.org/10.1186/s12859-019-3220-8
    ProtTrans models are used in this publication: https://doi.org/10.1101/2020.07.12.199554

    :param model: either "bert_from_publication" or "seqvec_from_publication". Used to download files
    """

    check_required(kwargs, ['embeddings_file', 'mapping_file', 'remapped_sequences_file'])
    result_kwargs = deepcopy(kwargs)
    file_manager = get_file_manager(**kwargs)

    # Download the checkpoint files if needed
    for file in BasicAnnotationExtractor.necessary_files:
        if not result_kwargs.get(file):
            result_kwargs[file] = get_model_file(model=f'{model}_annotations_extractors', file=file)

    annotation_extractor = BasicAnnotationExtractor(model, **result_kwargs)

    # mapping file will be needed for protein-wide annotations
    mapping_file = read_csv(result_kwargs['mapping_file'], index_col=0)

    # Try to create final files (if this fails, now is better than later
    DSSP3_predictions_file_path = file_manager.create_file(result_kwargs.get('prefix'),
                                                           result_kwargs.get('stage_name'),
                                                           'DSSP3_predictions_file',
                                                           extension='.fasta')
    result_kwargs['DSSP3_predictions_file'] = DSSP3_predictions_file_path
    DSSP8_predictions_file_path = file_manager.create_file(result_kwargs.get('prefix'),
                                                           result_kwargs.get('stage_name'),
                                                           'DSSP8_predictions_file',
                                                           extension='.fasta')
    result_kwargs['DSSP8_predictions_file'] = DSSP8_predictions_file_path
    disorder_predictions_file_path = file_manager.create_file(result_kwargs.get('prefix'),
                                                              result_kwargs.get('stage_name'),
                                                              'disorder_predictions_file',
                                                              extension='.fasta')
    result_kwargs['disorder_predictions_file'] = disorder_predictions_file_path
    per_sequence_predictions_file_path = file_manager.create_file(result_kwargs.get('prefix'),
                                                                  result_kwargs.get('stage_name'),
                                                                  'per_sequence_predictions_file',
                                                                  extension='.csv')
    result_kwargs['per_sequence_predictions_file'] = per_sequence_predictions_file_path

    # Create sequence containers
    DSSP3_sequences = list()
    DSSP8_sequences = list()
    disorder_sequences = list()

    with h5py.File(result_kwargs['embeddings_file'], 'r') as embedding_file:
        for protein_sequence in read_fasta(result_kwargs['remapped_sequences_file']):

            # Per-AA annotations: DSSP3, DSSP8 and disorder
            embedding = np.array(embedding_file[protein_sequence.id])

            annotations = annotation_extractor.get_annotations(embedding)

            DSSP3_sequence = deepcopy(protein_sequence)
            DSSP3_sequence.seq = Seq(convert_list_of_enum_to_string(annotations.DSSP3))
            DSSP3_sequences.append(DSSP3_sequence)

            DSSP8_sequence = deepcopy(protein_sequence)
            DSSP8_sequence.seq = Seq(convert_list_of_enum_to_string(annotations.DSSP8))
            DSSP8_sequences.append(DSSP8_sequence)

            disorder_sequence = deepcopy(protein_sequence)
            disorder_sequence.seq = Seq(convert_list_of_enum_to_string(annotations.disorder))
            disorder_sequences.append(disorder_sequence)

            # Per-sequence annotations, e.g. subcell loc & membrane boundness
            mapping_file.at[protein_sequence.id, 'subcellular_location'] = annotations.localization.value
            mapping_file.at[protein_sequence.id, 'membrane_or_soluble'] = annotations.membrane.value

    # Write files
    mapping_file.to_csv(per_sequence_predictions_file_path)
    write_fasta_file(DSSP3_sequences, DSSP3_predictions_file_path)
    write_fasta_file(DSSP8_sequences, DSSP8_predictions_file_path)
    write_fasta_file(disorder_sequences, disorder_predictions_file_path)

    return result_kwargs


# list of available extraction protocols
PROTOCOLS = {
    "bert_from_publication": bert_from_publication,
    "seqvec_from_publication": seqvec_from_publication,
    "unsupervised": unsupervised
}


def run(**kwargs):
    """
    Run embedding protocol

    Parameters
    ----------
    kwargs arguments (* denotes optional):
        prefix: Output prefix for all generated files
        protocol: Which protocol to use
        stage_name: The stage name

    Returns
    -------
    Dictionary with results of stage
    """
    check_required(kwargs, ['protocol', 'prefix', 'stage_name'])

    if kwargs["protocol"] not in PROTOCOLS:
        raise InvalidParameterError(
            "Invalid protocol selection: " +
            "{}. Valid protocols are: {}".format(
                kwargs["protocol"], ", ".join(PROTOCOLS.keys())
            )
        )

    return PROTOCOLS[kwargs["protocol"]](**kwargs)



import logging
import re
from itertools import zip_longest
from pathlib import Path
from typing import List, Generator

import torch
from numpy import ndarray
from transformers import T5Tokenizer, T5Model

from bio_embeddings.embed.embedder_interfaces import EmbedderWithFallback

logger = logging.getLogger(__name__)


class ProtTransT5BFDEmbedder(EmbedderWithFallback):
    """Encoder of the ProtTrans T5 BFD model

    Note that this model alone takes 13GB, so you need a GPU with a lot of memory.
    """

    _model: T5Model
    _decoder: bool = False
    name = "prottrans_t5_bfd"
    embedding_dimension = 1024
    number_of_layers = 1
    _necessary_directories = ["model_directory"]

    def __init__(self, **kwargs):
        """
        Initialize T5 embedder.

        :param model_directory:
        """
        super().__init__(**kwargs)

        self._model_directory = self._options["model_directory"]
        # Until we know whether we need the decoder, let's keep it here as an undocumented option.
        # Should the need arise we can just split this class in to an encoder and a decoder subclass
        # by setting one subclass to _decoder=True and the other to _decoder=False
        self._decoder = self._options.get("decoder", False)

        # make model
        self._model = T5Model.from_pretrained(self._model_directory)
        self._model = self._model.eval().to(self._device)
        self._model_fallback = None
        self._tokenizer = T5Tokenizer(
            str(Path(self._model_directory).joinpath("spiece.model")),
            do_lower_case=False,
        )

    def _get_fallback_model(self) -> T5Model:
        """ Returns the CPU model """
        if not self._model_fallback:
            self._model_fallback = T5Model.from_pretrained(self._model_directory).eval()
        return self._model_fallback

    def _embed_batch_impl(
        self, batch: List[str], model: T5Model
    ) -> Generator[ndarray, None, None]:
        seq_lens = [len(seq) for seq in batch]
        # Remove rare amino acids
        batch = [re.sub(r"[UZOB]", "X", sequence) for sequence in batch]
        # transformers needs spaces between the amino acids
        batch = [" ".join(list(seq)) for seq in batch]

        ids = self._tokenizer.batch_encode_plus(
            batch, add_special_tokens=True, pad_to_max_length=True
        )

        tokenized_sequences = torch.tensor(ids["input_ids"]).to(model.device)
        attention_mask = torch.tensor(ids["attention_mask"]).to(model.device)

        with torch.no_grad():
            embeddings = model(
                input_ids=tokenized_sequences,
                attention_mask=attention_mask,
                decoder_input_ids=tokenized_sequences,
            )

        # See comment in __init__
        if self._decoder:
            embeddings = embeddings[0].cpu().numpy()
        else:
            embeddings = embeddings[2].cpu().numpy()

        for seq_num, seq_len in zip_longest(range(len(embeddings)), seq_lens):
            # slice off last position (special token)
            embedding = embeddings[seq_num][:seq_len]
            assert (
                seq_len == embedding.shape[0]
            ), f"Sequence length mismatch: {seq_len} vs {embedding.shape[0]}"

            yield embedding

    @staticmethod
    def reduce_per_protein(embedding):
        return embedding.mean(axis=0)

    def embed(self, sequence: str) -> ndarray:
        [embedding] = self.embed_batch([sequence])
        return embedding



class NoEmbeddingException(Exception):
    """
    Exception to handle the case no embedding has been computed, and one requests features.
    """


class MissingParameterError(Exception):
    """
    Exception for missing parameters
    """


class InvalidParameterError(Exception):
    """
    Exception for invalid parameter settings
    """


class SequenceTooLongException(Exception):
    """
    Exception when trying to embed sequences longer then the limit
    """


class MD5ClashException(Exception):
    """
    When remapping sequences from a fasta file, if there is an MD5 clash, this will stop the execution
    """


class TooFewComponentsException(InvalidParameterError):
    """
    Thrown when n_components is nonsensical (e.g. < 2)
    """


class ConversionUniqueMismatch(Exception):
    """
    Thrown when trying to remap using a mapping file which doesn't have as many uniuqye original_ids as md5 hashes
    """


class UnrecognizedEmbeddingError(Exception):
    """
    Thrown when trying to access embeddings for sequences which have no embedding.
    """


class InvalidAnnotationFileError(Exception):
    """
    Thrown when an annotation file contains invalid values
    """


from sklearn.manifold import TSNE


def tsne_reduce(embeddings, **kwargs):
    """Wrapper around :meth:`sklearn.manifold.TSNE` with defaults for bio_embeddings"""
    tsne_params = dict()

    tsne_params['n_components'] = kwargs.get('n_components', 3)
    tsne_params['perplexity'] = kwargs.get('perplexity', 6)
    tsne_params['random_state'] = kwargs.get('random_state', 420)
    tsne_params['n_iter'] = kwargs.get('n_iter', 15000)
    tsne_params['verbose'] = kwargs.get('verbose', 1)
    tsne_params['n_jobs'] = kwargs.get('n_jobs', -1)
    tsne_params['metric'] = kwargs.get('metric', 'cosine')

    transformed_embeddings = TSNE(**tsne_params).fit_transform(embeddings)

    return transformed_embeddings



import logging
import re
from itertools import zip_longest
from pathlib import Path
from typing import Optional, Generator, List

import torch
from numpy import ndarray
from transformers import XLNetModel, XLNetTokenizer

from bio_embeddings.embed.embedder_interfaces import EmbedderInterface

logger = logging.getLogger(__name__)


class ProtTransXLNetUniRef100Embedder(EmbedderInterface):
    """ProtTrans-XLNet-UniRef100 Embedder (ProtXLNet)

    Elnaggar, Ahmed, et al. "ProtTrans: Towards Cracking the Language of Life's
    Code Through Self-Supervised Deep Learning and High Performance Computing."
    arXiv preprint arXiv:2007.06225 (2020). https://arxiv.org/abs/2007.06225
    """
    name = "prottrans_xlnet_uniref100"
    embedding_dimension = 1024
    number_of_layers = 1
    _model: XLNetModel
    _model_fallback: Optional[XLNetModel]
    _necessary_directories = ["model_directory"]

    def __init__(self, **kwargs):
        """
        Initialize XLNet embedder.

        :param model_directory:
        """
        super().__init__(**kwargs)

        # Get file locations from kwargs
        self.model_directory = self._options["model_directory"]

        # 512 is from https://github.com/agemagician/ProtTrans/blob/master/Embedding/PyTorch/Advanced/ProtXLNet.ipynb
        self._model = (
            XLNetModel.from_pretrained(self.model_directory, mem_len=512)
                .to(self._device)
                .eval()
        )
        self._model_fallback = None

        # sentence piece model
        # A standard text tokenizer which creates the input for NNs trained on text.
        # This one is just indexing single amino acids because we only have words of L=1.
        spm_model = str(Path(self.model_directory).joinpath("spm_model.model"))
        self._tokenizer = XLNetTokenizer.from_pretrained(spm_model, do_lower_case=False)

    def embed(self, sequence: str) -> ndarray:
        [embedding] = self.embed_batch([sequence])
        return embedding

    def embed_batch(self, batch: List[str]) -> Generator[ndarray, None, None]:
        seq_lens = [len(seq) for seq in batch]
        # transformers needs spaces between the amino acids
        batch = [" ".join(list(seq)) for seq in batch]
        # Remove rare amino acids
        batch = [re.sub(r"[UZOBX]", "<unk>", sequence) for sequence in batch]

        ids = self._tokenizer.batch_encode_plus(
            batch, add_special_tokens=True, pad_to_max_length=True
        )

        tokenized_sequences = torch.tensor(ids["input_ids"]).to(self._model.device)
        attention_mask = torch.tensor(ids["attention_mask"]).to(self._model.device)

        with torch.no_grad():
            embeddings, memory = self._model(
                input_ids=tokenized_sequences, attention_mask=attention_mask, mems=None
            )

        embeddings = embeddings.cpu().numpy()

        for seq_num, seq_len in zip_longest(range(len(embeddings)), seq_lens):
            attention_len = (attention_mask[seq_num] == 1).sum()
            padded_seq_len = len(attention_mask[seq_num])
            embedding = embeddings[seq_num][padded_seq_len - attention_len:padded_seq_len - 2]
            assert (
                    seq_len == embedding.shape[0]
            ), f"Sequence length mismatch: {seq_len} vs {embedding.shape[0]}"
            yield embedding

    @staticmethod
    def reduce_per_protein(embedding):
        return embedding.mean(axis=0)



"""Methods for predicting properties of proteins, both on a per-residue and
per-protein level, including supervised (pre-trained) and unsupervised (nearest
neighbour search) methods
"""

from bio_embeddings.extract.unsupervised_utilities import (
    pairwise_distance_matrix_from_embeddings_and_annotations,
    get_k_nearest_neighbours,
)
from bio_embeddings.extract.basic import BasicAnnotationExtractor

__all__ = [
    "BasicAnnotationExtractor",
    "get_k_nearest_neighbours",
    "pairwise_distance_matrix_from_embeddings_and_annotations",
]



import numpy as np
from gensim.models.keyedvectors import KeyedVectors

from bio_embeddings.embed.embedder_interfaces import EmbedderInterface


class FastTextEmbedder(EmbedderInterface):
    name = "fasttext"
    embedding_dimension = 1024
    number_of_layers = 1
    _necessary_files = ["model_file"]

    def __init__(self, **kwargs):
        """

        :param model_file: path of model file. If not supplied, will be downloaded.
        """
        super().__init__(**kwargs)

        self._model_file = self._options.get("model_file")

        self._model = KeyedVectors.load_word2vec_format(self._model_file, binary=False)
        self._vector_size = 512
        self._zero_vector = np.zeros(self._vector_size, dtype=np.float32)
        self._window_size = 3

    def embed(self, sequence):
        # pad sequence with special character (only 3-mers are considered)
        padded_sequence = "-" + sequence + "-"

        # container
        embedding = np.zeros((len(sequence), self._vector_size), dtype=np.float32)

        # for each aa in the sequence, retrieve k-mer
        for index in range(len(padded_sequence)):
            try:
                k_mer = "".join(padded_sequence[index : index + self._window_size])
                embedding[index, :] = self._get_kmer_representation(k_mer)
            # end of sequence reached
            except IndexError:
                return embedding

    def _get_kmer_representation(self, k_mer):
        # try to retrieve embedding for k-mer
        try:
            return self._model[k_mer]
        # in case of padded or out-of-vocab character
        except KeyError:
            # if single AA was not part of corpus (or no AA)
            if len(k_mer) <= 1:
                return self._zero_vector
            # handle border cases at start/end of seq
            elif "-" in k_mer:
                idx_center = int(len(k_mer) / 2)
                return self._get_kmer_representation(k_mer[idx_center])

    @staticmethod
    def reduce_per_protein(embedding):
        # TODO
        raise NotImplementedError



"""
The functionality of bio_embeddings is split into 5 different modules

.. autosummary::
   bio_embeddings.embed
   bio_embeddings.extract
   bio_embeddings.project
   bio_embeddings.utilities
   bio_embeddings.visualize
"""



from typing import List, Generator, Union

import torch
from esm.pretrained import load_model_and_alphabet_local
from numpy import ndarray

from bio_embeddings.embed import EmbedderInterface


class ESMEmbedder(EmbedderInterface):
    """ESM-1 Embedder (Note: This is not ESM-1b)

    Biological structure and function emerge from scaling unsupervised learning to 250 million protein sequences

    Rives, Alexander, et al. "Biological structure and function emerge from
    scaling unsupervised learning to 250 million protein sequences."
    bioRxiv (2019): 622803. https://doi.org/10.1101/622803
    """

    name = "esm"
    embedding_dimension = 1280
    number_of_layers = 1  # Following ESM, we only consider layer 34

    _necessary_files = ["model_file"]

    def __init__(self, device: Union[None, str, torch.device] = None, **kwargs):
        super().__init__(device, **kwargs)

        model, alphabet = load_model_and_alphabet_local(self._options["model_file"])

        self._model = model.to(self._device)
        self._batch_converter = alphabet.get_batch_converter()

    def embed(self, sequence: str) -> ndarray:
        [embedding] = self.embed_batch([sequence])
        return embedding

    def embed_batch(self, batch: List[str]) -> Generator[ndarray, None, None]:
        """See https://github.com/facebookresearch/esm/blob/master/README.rst#quickstart"""
        data = [(str(pos), sequence) for pos, sequence in enumerate(batch)]
        batch_labels, batch_strs, batch_tokens = self._batch_converter(data)

        with torch.no_grad():
            results = self._model(batch_tokens.to(self._device), repr_layers=[34])
        token_embeddings = results["representations"][34]

        # Generate per-sequence embeddings via averaging
        # NOTE: token 0 is always a beginning-of-sequence token, so the first residue is token 1.
        for i, (_, seq) in enumerate(data):
            yield token_embeddings[i, 1 : len(seq) + 1].cpu().numpy()

    @staticmethod
    def reduce_per_protein(embedding: ndarray) -> ndarray:
        return embedding.mean(0)



import logging
import os
import string
from copy import deepcopy
from datetime import datetime
from typing import Dict, Callable, Optional, Any
from urllib import request

import importlib_metadata
from importlib_metadata import PackageNotFoundError

from bio_embeddings.embed.pipeline import run as run_embed
from bio_embeddings.extract.pipeline import run as run_extract
from bio_embeddings.project.pipeline import run as run_project
from bio_embeddings.utilities import get_file_manager, read_fasta, reindex_sequences, write_fasta_file, \
    check_required, MD5ClashException
from bio_embeddings.utilities.config import read_config_file, write_config_file
from bio_embeddings.utilities.filemanagers import FileManagerInterface
from bio_embeddings.utilities.remote_file_retriever import TqdmUpTo
from bio_embeddings.visualize.pipeline import run as run_visualize

logger = logging.getLogger(__name__)

_STAGES = {
    "embed": run_embed,
    "project": run_project,
    "visualize": run_visualize,
    "extract": run_extract
}

_IN_CONFIG_NAME = "input_parameters_file"
_OUT_CONFIG_NAME = "ouput_parameters_file"


def _valid_file(file_path):
    """
    Verify if a file exists and is not empty.
    Parameters
    ----------
    file_path : str
        Path to file to check
    Returns
    -------
    bool
        True if file exists and is non-zero size,
        False otherwise.
    """
    try:
        return os.stat(file_path).st_size > 0
    except (OSError, TypeError):
        # catch TypeError for nonsense paths, e.g. None
        return False


def _process_fasta_file(**kwargs):
    """
    Will assign MD5 hash as ID if no if provided for a sequence.
    """
    result_kwargs = deepcopy(kwargs)
    file_manager = get_file_manager(**kwargs)

    sequences = read_fasta(kwargs['sequences_file'])

    # Sanity check the fasta file to avoid nonsense and/or crashes by the embedders
    letters = set(string.ascii_letters)
    for entry in sequences:
        illegal = sorted(set(entry.seq) - letters)
        if illegal:
            formatted = "'" + "', '".join(illegal) + "'"
            raise ValueError(
                f"The entry '{entry.name}' in {kwargs['sequences_file']} contains the characters {formatted}, "
                f"while only single letter code is allowed "
                f"(https://en.wikipedia.org/wiki/Amino_acid#Table_of_standard_amino_acid_abbreviations_and_properties)."
            )
        # This is a warning due to the inconsistent handling between different embedders
        if not str(entry.seq).isupper():
            logger.warning(
                f"The entry '{entry.name}' in {kwargs['sequences_file']} contains lower case amino acids. "
                f"Lower case letters are uninterpretable by most language models, "
                f"and their embedding will be nonesensical. "
                f"Protein LMs available through bio_embeddings have been trained on upper case, "
                f"single letter code sequence representations only "
                f"(https://en.wikipedia.org/wiki/Amino_acid#Table_of_standard_amino_acid_abbreviations_and_properties)."
            )

    sequences_file_path = file_manager.create_file(kwargs.get('prefix'), None, 'sequences_file',
                                                   extension='.fasta')
    write_fasta_file(sequences, sequences_file_path)

    result_kwargs['sequences_file'] = sequences_file_path

    # Remap using sequence position rather than md5 hash -- not encouraged!
    result_kwargs['simple_remapping'] = result_kwargs.get('simple_remapping', False)

    mapping = reindex_sequences(sequences, simple=result_kwargs['simple_remapping'])

    # Check if there's the same MD5 index twice. This most likely indicates 100% sequence identity.
    # Throw an error for MD5 hash clashes!
    if mapping.index.has_duplicates:
        raise MD5ClashException("There is at least one MD5 hash clash.\n"
                                "This most likely indicates there are multiple identical sequences in your FASTA file.\n"
                                "MD5 hashes are used to remap sequence identifiers from the input FASTA.\n"
                                "This error exists to prevent wasting resources (computing the same embedding twice).\n"
                                "There's a (very) low probability of this indicating a real MD5 clash.\n\n"
                                "If you are sure there are no identical sequences in your set, please open an issue at "
                                "https://github.com/sacdallago/bio_embeddings/issues . "
                                "Otherwise, use cd-hit to reduce your input FASTA to exclude identical sequences!")

    mapping_file_path = file_manager.create_file(kwargs.get('prefix'), None, 'mapping_file', extension='.csv')
    remapped_sequence_file_path = file_manager.create_file(kwargs.get('prefix'), None, 'remapped_sequences_file',
                                                           extension='.fasta')

    write_fasta_file(sequences, remapped_sequence_file_path)
    mapping.to_csv(mapping_file_path)

    result_kwargs['mapping_file'] = mapping_file_path
    result_kwargs['remapped_sequences_file'] = remapped_sequence_file_path

    return result_kwargs


def _null_function(config: Dict) -> None:
    pass


def download_files_for_stage(
    stage_parameters: Dict[str, Any],
    file_manager: FileManagerInterface,
    prefix: str,
    stage_name: Optional[str] = None,
) -> Dict[str, Any]:
    """Download files given as url

    We don't actually check whether a given option actually takes a path, e.g.
    one could specify `simple_remapping: https://google.com` and we'd download
    google.com. It will however still be evident for the user that the error
    lies in `simple_remapping` even though we replaced the value
    """
    for key in stage_parameters:
        if isinstance(stage_parameters[key], str) and (
            stage_parameters[key].startswith("http://")
            or stage_parameters[key].startswith("https://")
            or stage_parameters[key].startswith("ftp://")
        ):
            filename = file_manager.create_file(prefix, stage_name, key)
            logger.info(f"Downloading {stage_parameters[key]} to {filename}")
            desc = stage_parameters[key].split("/")[-1]
            with TqdmUpTo(unit="B", unit_scale=True, miniters=1, desc=desc) as tqdm:
                request.urlretrieve(
                    stage_parameters[key], filename=filename, reporthook=tqdm.update_to
                )
            stage_parameters[key] = filename
    return stage_parameters


def execute_pipeline_from_config(config: Dict,
                                 post_stage: Callable[[Dict], None] = _null_function,
                                 **kwargs) -> Dict:
    original_config = deepcopy(config)

    check_required(
        config,
        ["global"]
    )

    # !! pop = remove from config!
    global_parameters = config.pop('global')

    check_required(
        global_parameters,
        ["prefix", "sequences_file"]
    )

    file_manager = get_file_manager(**global_parameters)

    # Make sure prefix exists
    prefix = global_parameters['prefix']

    # If prefix already exists
    if file_manager.exists(prefix):
        if not kwargs.get('overwrite'):
            raise FileExistsError("The prefix already exists & no overwrite option has been set.\n"
                                  "Either set --overwrite, or move data from the prefix.\n"
                                  "Prefix: {}".format(prefix))
    else:
        # create the prefix
        file_manager.create_prefix(prefix)

    # Copy original config to prefix
    global_in = file_manager.create_file(prefix, None, _IN_CONFIG_NAME, extension='.yml')
    write_config_file(global_in, original_config)

    # This downloads sequences_file if required
    download_files_for_stage(global_parameters, file_manager, prefix)

    global_parameters = _process_fasta_file(**global_parameters)

    for stage_name in config:
        stage_parameters = config[stage_name]

        check_required(
            stage_parameters,
            ["protocol", "type"]
        )

        stage_type = stage_parameters['type']
        stage_runnable = _STAGES.get(stage_type)

        if not stage_runnable:
            raise Exception("No type defined, or invalid stage type defined: {}".format(stage_type))

        # Prepare to run stage
        stage_parameters['stage_name'] = stage_name
        file_manager.create_stage(prefix, stage_name)

        stage_parameters = download_files_for_stage(stage_parameters, file_manager, prefix, stage_name)

        stage_dependency = stage_parameters.get('depends_on')

        if stage_dependency:
            if stage_dependency not in config:
                raise Exception("Stage {} depends on {}, but dependency not found in config.".format(stage_name,
                                                                                                     stage_dependency))

            stage_dependency_parameters = config.get(stage_dependency)
            stage_parameters = {**global_parameters, **stage_dependency_parameters, **stage_parameters}
        else:
            stage_parameters = {**global_parameters, **stage_parameters}

        # Register start time
        start_time = datetime.now().astimezone()
        stage_parameters['start_time'] = str(start_time)

        stage_in = file_manager.create_file(prefix, stage_name, _IN_CONFIG_NAME, extension='.yml')
        write_config_file(stage_in, stage_parameters)

        stage_output_parameters = stage_runnable(**stage_parameters)

        # Register end time
        end_time = datetime.now().astimezone()
        stage_output_parameters['end_time'] = str(end_time)

        # Register elapsed time
        stage_output_parameters['elapsed_time'] = str(end_time - start_time)

        stage_out = file_manager.create_file(prefix, stage_name, _OUT_CONFIG_NAME, extension='.yml')
        write_config_file(stage_out, stage_output_parameters)

        # Store in global_out config for later retrieval (e.g. depends_on)
        config[stage_name] = stage_output_parameters

        # Execute post-stage function, if provided
        post_stage(stage_output_parameters)

    config['global'] = global_parameters

    try:
        config['global']['version'] = importlib_metadata.version("bio_embeddings")
    except PackageNotFoundError:
        pass  # :(

    global_out = file_manager.create_file(prefix, None, _OUT_CONFIG_NAME, extension='.yml')
    write_config_file(global_out, config)

    return config


def parse_config_file_and_execute_run(config_file_path: str, **kwargs):
    if not _valid_file(config_file_path):
        raise Exception("No config or invalid config was passed.")

    # read configuration and execute
    config = read_config_file(config_file_path)

    execute_pipeline_from_config(config, **kwargs)



"""
Most of this implementation is taken from
https://github.com/tbepler/protein-sequence-embedding-iclr2019/blob/3bb338bd70e2b7b97c733304d50cfcac9c35cb27/embed_sequences.py

---

Supporting torch > 1.3 was a bit tricky (https://github.com/tbepler/protein-sequence-embedding-iclr2019/issues/21).
Here's what I did:

First, download and unpack the weights to `pretrained_weights`.
```
wget http://bergerlab-downloads.csail.mit.edu/bepler-protein-sequence-embeddings-from-structure-iclr2019/pretrained_models.tar.gz
tar xf pretrained_models.tar.gz
```

Create a torch 1.3 virtualenv (with python 3.7). In this venv, run:

```shell_script
git clone https://github.com/tbepler/protein-sequence-embedding-iclr2019
cd protein-sequence-embedding-iclr2019
pip install Cython
python setup.py install
```

```python
import torch

model = torch.load("pretrained_models/ssa_L1_100d_lstm3x512_lm_i512_mb64_tau0.5_lambda0.1_p0.05_epoch100.sav")
# For some reason torch seems to have missed that in older versions, but requires it in newer ones
state_dict = model.state_dict()
state_dict["scop_predict.gap"] = torch.FloatTensor([-10])
torch.save(state_dict, "pretrained_models/ssa_L1_100d_lstm3x512_lm_i512_mb64_tau0.5_lambda0.1_p0.05_epoch100_updated.state_dict")
print(model)
```

I then switched back to my normal torch 1.5.1 environment and recreated the
printed model params as you can see in the __init__ function.
"""

from typing import Union

import torch
from bepler.alphabets import Uniprot21
from bepler.models.embedding import StackedRNN
from bepler.models.multitask import SCOPCM
from bepler.models.sequence import BiLM
from numpy import ndarray
from torch import nn

from bio_embeddings.embed import EmbedderInterface


def _unstack_lstm(lstm, device: torch.device):
    in_size = lstm.input_size
    hidden_dim = lstm.hidden_size
    layers = []
    for i in range(lstm.num_layers):
        layer = nn.LSTM(in_size, hidden_dim, batch_first=True, bidirectional=True)
        layer.to(device)

        attributes = ["weight_ih_l", "weight_hh_l", "bias_ih_l", "bias_hh_l"]
        for attr in attributes:
            dest = attr + "0"
            src = attr + str(i)
            getattr(layer, dest).data[:] = getattr(lstm, src)

            dest = attr + "0_reverse"
            src = attr + str(i) + "_reverse"
            getattr(layer, dest).data[:] = getattr(lstm, src)
        layer.flatten_parameters()
        layers.append(layer)
        in_size = 2 * hidden_dim
    return layers


class BeplerEmbedder(EmbedderInterface):
    """Bepler Embedder

    Bepler, Tristan, and Bonnie Berger. "Learning protein sequence embeddings using information from structure."
    arXiv preprint arXiv:1902.08661 (2019).
    """

    name = "bepler"
    embedding_dimension = 121  # 100 + len(self.alphabet)
    number_of_layers = 1

    # This is derived from ssa_L1_100d_lstm3x512_lm_i512_mb64_tau0.5_lambda0.1_p0.05_epoch100.sav
    # See text at the top of the file
    _necessary_files = ["model_file"]

    def __init__(self, device: Union[None, str, torch.device] = None, **kwargs):
        super().__init__(device, **kwargs)
        self.alphabet = Uniprot21()

        # These parameters are part of the model, but we can't load them if we
        # use the state dict
        hidden = 512
        out = 100

        lm = BiLM(
            len(self.alphabet) + 1,
            len(self.alphabet),
            len(self.alphabet),
            hidden * 2,
            2,
        )
        embedding = StackedRNN(
            len(self.alphabet),
            hidden,
            hidden,
            out,
            nlayers=3,
            dropout=0,
            lm=lm,
        )

        self.model = SCOPCM(embedding)
        self.model.load_state_dict(torch.load(self._options["model_file"]))
        self.model = self.model.eval().to(self._device)

        self.lstm_stack = _unstack_lstm(self.model.embedding.rnn, self._device)

    def embed(self, sequence: str) -> ndarray:
        x = sequence.upper().encode()
        # convert to alphabet index
        x = self.alphabet.encode(x)
        x = torch.from_numpy(x).to(self._device)
        # embed the sequence
        with torch.no_grad():
            x = x.long().unsqueeze(0)
            zs = []
            # noinspection PyUnresolvedReferences
            x_onehot = x.new(x.size(0), x.size(1), 21).float().zero_()
            x_onehot.scatter_(2, x.unsqueeze(2), 1)
            zs.append(x_onehot)
            h = self.model.embedding.embed(x)
            for lstm in self.lstm_stack:
                h, _ = lstm(h)
            h = self.model.embedding.proj(h.squeeze(0)).unsqueeze(0)
            zs.append(h)
            z1 = torch.cat(zs, 2)
            z = z1
            return z.squeeze(0).cpu().numpy()

    @staticmethod
    def reduce_per_protein(embedding: ndarray) -> ndarray:
        return embedding.mean(axis=0)



import logging
import os
import tempfile
import zipfile
from pathlib import Path
from typing import Dict, Optional
from urllib import request

from appdirs import user_cache_dir
from tqdm import tqdm

from bio_embeddings.utilities.config import read_config_file

_module_dir: Path = Path(os.path.dirname(os.path.abspath(__file__)))
_defaults: Dict[str, Dict[str, str]] = read_config_file(_module_dir / "defaults.yml")

logger = logging.getLogger(__name__)


class TqdmUpTo(tqdm):
    """Provides `update_to(n)` which uses `tqdm.update(delta_n)`."""

    def update_to(self, b=1, bsize=1, tsize=None):
        """
        b  : int, optional
            Number of blocks transferred so far [default: 1].
        bsize  : int, optional
            Size of each block (in tqdm units) [default: 1].
        tsize  : int, optional
            Total size (in tqdm units). If [default: None] remains unchanged.
        """
        if tsize is not None:
            self.total = tsize
        self.update(b * bsize - self.n)  # will also set self.n = b * bsize


def get_model_directories_from_zip(
    model: Optional[str] = None,
    directory: Optional[str] = None,
    overwrite_cache: bool = False,
) -> str:
    """If the specified asset directory for the model is in the user cache, returns the
    directory path, otherwise downloads the zipped directory, unpacks in the cache and
    returns the location"""
    cache_path = (
        Path(user_cache_dir("bio_embeddings")).joinpath(model).joinpath(directory)
    )
    if (
        not overwrite_cache
        and cache_path.is_dir()
        and len(list(cache_path.iterdir())) > 1
    ):
        logger.info(f"Loading {directory} for {model} from cache at '{cache_path}'")
        return str(cache_path)

    cache_path.mkdir(parents=True, exist_ok=True)
    url = _defaults.get(model, {}).get(directory)

    # Since the directory are not user provided, this must never happen
    assert url, f"Directory {directory} for {model} doesn't exist."

    with tempfile.NamedTemporaryFile() as f:
        file_name = f.name

        logger.info(
            "Downloading {} for {} and storing in '{}'.".format(
                "model_folder_zip", model, file_name
            )
        )

        with TqdmUpTo(
            unit="B", unit_scale=True, miniters=1, desc=url.split("/")[-1]
        ) as t:
            request.urlretrieve(url, filename=file_name, reporthook=t.update_to)

        logger.info(
            "Unzipping {} for {} and storing in '{}'.".format(
                file_name, model, cache_path
            )
        )

        with zipfile.ZipFile(file_name, "r") as zip_ref:
            zip_ref.extractall(cache_path)

    return str(cache_path)


def get_model_file(
    model: Optional[str] = None,
    file: Optional[str] = None,
    overwrite_cache: bool = False,
) -> str:
    """If the specified asset for the model is in the user cache, returns the
    location, otherwise downloads the file to cache and returns the location"""
    cache_path = Path(user_cache_dir("bio_embeddings")).joinpath(model).joinpath(file)
    if not overwrite_cache and cache_path.is_file():
        logger.info(f"Loading {file} for {model} from cache at '{cache_path}'")
        return str(cache_path)

    cache_path.parent.mkdir(exist_ok=True, parents=True)
    url = _defaults.get(model, {}).get(file)

    # Since the files are not user provided, this must never happen
    assert url, f"File {file} for {model} doesn't exist."

    logger.info(f"Downloading {file} for {model} and storing it in '{cache_path}'")

    with TqdmUpTo(unit="B", unit_scale=True, miniters=1, desc=url.split("/")[-1]) as t:
        request.urlretrieve(url, filename=cache_path, reporthook=t.update_to)

    return str(cache_path)



import gc
import subprocess
import sys
from argparse import ArgumentParser
from time import sleep

import torch

from bio_embeddings.embed import name_to_embedder


def embedder_round_trip(name: str) -> int:
    """This function should tell us the size of an embedder and not leak memory"""
    memory_allocated_a = torch.cuda.memory_allocated(0)
    all_allocated_a = torch.cuda.memory_stats()["active.all.allocated"]
    # noinspection PyUnusedLocal
    embedder = name_to_embedder[name]()
    embedder.embed_many(["PROTEIN", "SEQWENCE"])
    memory_allocated_b = torch.cuda.memory_allocated(0)
    all_allocated_b = torch.cuda.memory_stats()["active.all.allocated"]
    del embedder
    gc.collect()
    torch.cuda.empty_cache()
    memory_allocated_c = torch.cuda.memory_allocated(0)
    all_allocated_c = torch.cuda.memory_stats()["active.all.allocated"]
    print(name, memory_allocated_a, memory_allocated_b, memory_allocated_c)
    print(name, all_allocated_a, all_allocated_b, all_allocated_c)
    return memory_allocated_b - memory_allocated_a


def main():
    parser = ArgumentParser()
    parser.add_argument("embedder", help="Lowercase name of the embedder or all")
    args = parser.parse_args()
    if args.embedder == "all":
        for name in name_to_embedder:
            subprocess.check_call([sys.executable, __file__, name])
    elif args.embedder == "all-keep":
        ram_by_name = {}
        for name in name_to_embedder:
            print(name)
            ram_by_name[name] = embedder_round_trip(name)
        print(ram_by_name)
        sleep(100) # Let me read nvidia-smi
    else:
        embedder_round_trip(args.embedder)


if __name__ == "__main__":
    main()



import abc


class FileManagerInterface(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def __init__(self):
        pass

    @abc.abstractmethod
    def create_prefix(self, prefix) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def create_stage(self, prefix, stage) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def create_file(self, prefix, stage, file_name, extension=None) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def create_directory(self, prefix, stage, directory_name) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def get_file(self, prefix, stage, file_name, extension=None) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def exists(self, prefix, stage=None, file_name=None, extension=None) -> bool:
        raise NotImplementedError



"""
Configuration handling

Authors:
  Thomas A. Hopf
"""
from pathlib import Path
from typing import Union

import ruamel.yaml as yaml
from ruamel.yaml.parser import ParserError
from ruamel.yaml.comments import CommentedBase

from bio_embeddings.utilities.exceptions import InvalidParameterError


def parse_config(config_str: str, preserve_order: bool = True) -> dict:
    """
    Parse a configuration string

    Parameters
    ----------
    config_str : str
        Configuration to be parsed
    preserve_order : bool, optional (default: False)
        Preserve formatting of input configuration
        string

    Returns
    -------
    dict
        Configuration dictionary
    """
    try:
        if preserve_order:
            return yaml.load(config_str, Loader=yaml.RoundTripLoader)
        else:
            return yaml.safe_load(config_str)
    except ParserError as e:
        raise InvalidParameterError(
            "Could not parse input configuration. "
            "Formatting mistake in config file? "
            "See ParserError above for details."
        ) from e


def read_config_file(config_path: Union[str, Path], preserve_order: bool = True) -> dict:
    """
    Read config from path to file.

    :param config_path: path to .yml config file
    :param preserve_order:
    :return:
    """
    with open(config_path, "r") as f:
        return parse_config(f.read(), preserve_order)


def write_config_file(out_filename: str, config: dict) -> None:
    """
    Save configuration data structure in YAML file.

    Parameters
    ----------
    out_filename : str
        Filename of output file
    config : dict
        Config data that will be written to file
    """
    if isinstance(config, CommentedBase):
        dumper = yaml.RoundTripDumper
    else:
        dumper = yaml.Dumper

    with open(out_filename, "w") as f:
        f.write(
            yaml.dump(config, Dumper=dumper, default_flow_style=False)
        )



"""Language models to translate amino acid sequences into vector representations

All language models implement the :class:`EmbedderInterface`. You can embed a single
sequences with :meth:`EmbedderInterface.embed` or a list of
sequences with the :meth:`EmbedderInterface.embed_many` function. All except
CPCProt and UniRep generate per-residue embeddings, which you can summarize
into a fixed size per-protein embedding by calling
:meth:`EmbedderInterface.reduce_per_protein`. CPCProt only generates a
per-protein embedding (``reduce_per_protein`` does nothing), while UniRep
includes a start token, so the embedding is one longer than the protein.

Instead of using ``bio_embeddings[all]``, it's possible to only install
some embedders by selecting specific extras:

* ``allennlp``: seqvec
* ``transformers``: prottrans_albert_bfd, prottrans_bert_bfd, protrans_xlnet_uniref100, prottrans_t5_bfd
* ``jax``-unirep: unirep
* ``esm``: esm
* ``cpcprot``: cpcprot
* ``plus``: plus_rnn
"""

import logging
from typing import Dict, Type

from bio_embeddings.embed.embedder_interfaces import EmbedderInterface

logger = logging.getLogger(__name__)

name_to_embedder: Dict[str, Type[EmbedderInterface]] = {}
__all__ = ["EmbedderInterface"]

# Transformers
try:
    from bio_embeddings.embed.prottrans_albert_bfd_embedder import (
        ProtTransAlbertBFDEmbedder,
    )
    from bio_embeddings.embed.prottrans_bert_bfd_embedder import (
        ProtTransBertBFDEmbedder,
    )
    from bio_embeddings.embed.prottrans_xlnet_uniref100_embedder import (
        ProtTransXLNetUniRef100Embedder,
    )
    from bio_embeddings.embed.prottrans_t5_bfd_embedder import ProtTransT5BFDEmbedder

    name_to_embedder[ProtTransAlbertBFDEmbedder.name] = ProtTransAlbertBFDEmbedder
    name_to_embedder[ProtTransBertBFDEmbedder.name] = ProtTransBertBFDEmbedder
    name_to_embedder[
        ProtTransXLNetUniRef100Embedder.name
    ] = ProtTransXLNetUniRef100Embedder
    name_to_embedder[ProtTransT5BFDEmbedder.name] = ProtTransT5BFDEmbedder

    __all__.append(ProtTransAlbertBFDEmbedder.__name__)
    __all__.append(ProtTransBertBFDEmbedder.__name__)
    __all__.append(ProtTransXLNetUniRef100Embedder.__name__)
    __all__.append(ProtTransT5BFDEmbedder.__name__)
except ImportError:
    logger.debug(
        "transformers extra not installed, Bert, Albert and XLNet will not be available"
    )

# Elmo / SeqVec
try:
    from bio_embeddings.embed.seqvec_embedder import SeqVecEmbedder

    name_to_embedder[SeqVecEmbedder.name] = SeqVecEmbedder
    __all__.append(SeqVecEmbedder.__name__)
except ImportError:
    logger.debug("allennlp extra not installed, SeqVec will not be available")

if not name_to_embedder:
    logger.warning(
        "No extra is installed, so none of the context dependent embedders are available! "
        "Please run `pip install bio-embeddings[all]`!"
    )

# ESM
try:
    from bio_embeddings.embed.esm_embedder import ESMEmbedder

    name_to_embedder[ESMEmbedder.name] = ESMEmbedder
    __all__.append(ESMEmbedder.__name__)
except ImportError:
    logger.debug("esm extra is not installed, ESM will not be available")

# UniRep
try:
    from bio_embeddings.embed.unirep_embedder import UniRepEmbedder

    name_to_embedder[UniRepEmbedder.name] = UniRepEmbedder
    __all__.append(UniRepEmbedder.__name__)
except ImportError:
    logger.debug("unirep extra not installedm UniRep will not be available")

# CPCProt
try:
    from bio_embeddings.embed.cpcprot_embedder import CPCProtEmbedder

    name_to_embedder[CPCProtEmbedder.name] = CPCProtEmbedder
    __all__.append(CPCProtEmbedder.__name__)
except ImportError:
    logger.debug("cpcprot extra not installed, CPCProt will not be available")

# PLUS
try:
    from bio_embeddings.embed.plus_rnn_embedder import PLUSRNNEmbedder

    name_to_embedder[PLUSRNNEmbedder.name] = PLUSRNNEmbedder
    __all__.append(PLUSRNNEmbedder.__name__)
except ImportError:
    logger.debug("plus extra not installed, PLUSRNNEmbedder will not be available")

# Bepler - should always work
try:
    from bio_embeddings.embed.bepler_embedder import BeplerEmbedder

    name_to_embedder[BeplerEmbedder.name] = BeplerEmbedder
    __all__.append(BeplerEmbedder.__name__)
except ImportError:
    logger.debug("bepler extra not installed, PLUSRNNEmbedder will not be available")

# Unmaintained embedders
from bio_embeddings.embed.fasttext_embedder import FastTextEmbedder
from bio_embeddings.embed.glove_embedder import GloveEmbedder
from bio_embeddings.embed.word2vec_embedder import Word2VecEmbedder



import logging
from pathlib import Path

from transformers import BertModel, BertTokenizer

from bio_embeddings.embed.prottrans_base_embedder import ProtTransBertBaseEmbedder

logger = logging.getLogger(__name__)


class ProtTransBertBFDEmbedder(ProtTransBertBaseEmbedder):
    """ProtTrans-Bert-BFD Embedder (ProtBert-BFD)

    Elnaggar, Ahmed, et al. "ProtTrans: Towards Cracking the Language of Life's
    Code Through Self-Supervised Deep Learning and High Performance Computing."
    arXiv preprint arXiv:2007.06225 (2020). https://arxiv.org/abs/2007.06225
    """
    _model: BertModel
    name = "prottrans_bert_bfd"
    embedding_dimension = 1024
    number_of_layers = 1

    def __init__(self, **kwargs):
        """
        Initialize Bert embedder.

        :param model_directory:
        """
        super().__init__(**kwargs)

        self._model_directory = self._options["model_directory"]

        # make model
        self._model = BertModel.from_pretrained(self._model_directory)
        self._model = self._model.eval().to(self._device)
        self._model_fallback = None
        self._tokenizer = BertTokenizer(
            str(Path(self._model_directory) / "vocab.txt"), do_lower_case=False
        )

    def _get_fallback_model(self) -> BertModel:
        """ Returns the CPU model """
        if not self._model_fallback:
            self._model_fallback = BertModel.from_pretrained(
                self._model_directory
            ).eval()
        return self._model_fallback



import logging

from copy import deepcopy
from pandas import read_csv
from bio_embeddings.utilities import InvalidParameterError, check_required, get_file_manager, TooFewComponentsException
from bio_embeddings.visualize import render_3D_scatter_plotly, render_scatter_plotly, \
    save_plotly_figure_to_html

logger = logging.getLogger(__name__)


def plotly(**kwargs):
    result_kwargs = deepcopy(kwargs)
    file_manager = get_file_manager(**kwargs)

    # 2 or 3D plot? Usually, this is directly fetched from "project" stage via "depends_on"
    result_kwargs['n_components'] = kwargs.get('n_components', 3)
    if result_kwargs['n_components'] < 2:
        raise TooFewComponentsException(f"n_components is set to {result_kwargs['n_components']}. It should be >1.\n"
                                        f"If set to 2, will render 2D scatter plot.\n"
                                        f"If set to >3, will render 3D scatter plot.")

    # Get projected_embeddings_file containing x,y,z coordinates and identifiers
    projected_embeddings_file = read_csv(result_kwargs['projected_embeddings_file'], index_col=0)

    if result_kwargs.get('annotation_file'):

        annotation_file = read_csv(result_kwargs['annotation_file']).set_index('identifier')

        # Save a copy of the annotation file with index set to identifier
        input_annotation_file_path = file_manager.create_file(kwargs.get('prefix'),
                                                              result_kwargs.get('stage_name'),
                                                              'input_annotation_file',
                                                              extension='.csv')

        annotation_file.to_csv(input_annotation_file_path)

        # Merge annotation file and projected embedding file based on index or original id?
        result_kwargs['merge_via_index'] = result_kwargs.get('merge_via_index', False)

        # Display proteins with unknown annotation?
        result_kwargs['display_unknown'] = result_kwargs.get('display_unknown', True)

        if result_kwargs['merge_via_index']:
            if result_kwargs['display_unknown']:
                merged_annotation_file = annotation_file.join(projected_embeddings_file, how="outer")
                merged_annotation_file['label'].fillna('UNKNOWN', inplace=True)
            else:
                merged_annotation_file = annotation_file.join(projected_embeddings_file)
        else:
            if result_kwargs['display_unknown']:
                merged_annotation_file = annotation_file.join(projected_embeddings_file.set_index('original_id'), how="outer")
                merged_annotation_file['label'].fillna('UNKNOWN', inplace=True)
            else:
                merged_annotation_file = annotation_file.join(projected_embeddings_file.set_index('original_id'))

        merged_annotation_file_path = file_manager.create_file(kwargs.get('prefix'),
                                                               result_kwargs.get('stage_name'),
                                                               'merged_annotation_file',
                                                               extension='.csv')

        merged_annotation_file.to_csv(merged_annotation_file_path)
        result_kwargs['merged_annotation_file'] = merged_annotation_file_path
    else:
        projected_embeddings_file['label'] = 'UNKNOWN'

    if result_kwargs['n_components'] == 2:
        figure = render_scatter_plotly(projected_embeddings_file)
    else:
        figure = render_3D_scatter_plotly(projected_embeddings_file)

    plot_file_path = file_manager.create_file(kwargs.get('prefix'),
                                              result_kwargs.get('stage_name'),
                                              'plot_file',
                                              extension='.html')

    save_plotly_figure_to_html(figure, plot_file_path)
    result_kwargs['plot_file'] = plot_file_path

    return result_kwargs


# list of available projection protocols
PROTOCOLS = {
    "plotly": plotly,
}


def run(**kwargs):
    """
    Run visualize protocol

    Parameters
    ----------
    kwargs arguments (* denotes optional):
        projected_embeddings_file: A csv with columns: (index), original_id, x, y, z
        prefix: Output prefix for all generated files
        stage_name: The stage name
        protocol: Which plot to generate

    Returns
    -------
    Dictionary with results of stage
    """
    check_required(kwargs, ['protocol', 'prefix', 'stage_name', 'projected_embeddings_file'])

    if kwargs["protocol"] not in PROTOCOLS:
        raise InvalidParameterError(
            "Invalid protocol selection: " +
            "{}. Valid protocols are: {}".format(
                kwargs["protocol"], ", ".join(PROTOCOLS.keys())
            )
        )

    return PROTOCOLS[kwargs["protocol"]](**kwargs)


"""Display t-SNE or UMAP projections from the project stage"""

from bio_embeddings.visualize.plotly_plots import (
    render_3D_scatter_plotly,
    render_scatter_plotly,
    save_plotly_figure_to_html,
)

__all__ = [
    "render_3D_scatter_plotly",
    "render_scatter_plotly",
    "save_plotly_figure_to_html",
]



from typing import List, Generator, Union

import numpy
import torch
from CPCProt import CPCProtModel, CPCProtEmbedding, CPCProtConfig
from CPCProt.model.cpcprot import DEFAULT_CONFIG
from CPCProt.tokenizer import Tokenizer
from numpy import ndarray

from bio_embeddings.embed import EmbedderInterface


class CPCProtEmbedder(EmbedderInterface):
    """CPCProt Embedder

    Lu, Amy X., et al. "Self-supervised contrastive learning of protein
    representations by mutual information maximization." bioRxiv (2020).
    https://doi.org/10.1101/2020.09.04.283929
    """

    name = "cpcprot"
    embedding_dimension = 512
    number_of_layers = 1

    _necessary_files = ["model_file"]

    def __init__(self, device: Union[None, str, torch.device] = None, **kwargs):
        super().__init__(device, **kwargs)
        self.tokenizer = Tokenizer(vocab="iupac")
        # If we don't do this here, CPCProtModel will end up on the gpu if one is
        # available, even if we passed the cpu as device.
        # Afaik this is the best way to derive from DEFAULT_CONFIG
        dict_cfg = DEFAULT_CONFIG.to_dict()
        dict_cfg["use_cuda"] = self._device.type == "cuda"
        raw_model = CPCProtModel(cfg=CPCProtConfig.from_dict(dict_cfg)).to(self._device)
        state_dict = dict(
            torch.load(self._options["model_file"], map_location=self._device)
        )
        for i in list(state_dict.keys()):
            if i.startswith("module."):
                state_dict[i[7:]] = state_dict[i]
                del state_dict[i]
        raw_model.load_state_dict(state_dict)
        self._model = CPCProtEmbedding(raw_model.to(self._device).eval())

    def embed(self, sequence: str) -> ndarray:
        [embedding] = self.embed_batch([sequence])
        return embedding

    def embed_batch(self, batch: List[str]) -> Generator[ndarray, None, None]:
        """See https://github.com/amyxlu/CPCProt/blob/df1ad1118544ed349b5e711207660a7c205b3128/embed_fasta.py"""
        encoded = [numpy.array(self.tokenizer.encode(sequence)) for sequence in batch]
        # 11 is the minimum patch size, so we need to zero-pad shorter sequences
        pad_length = max(max([i.shape[0] for i in encoded]), 11)
        padded = [numpy.pad(i, (0, pad_length - i.shape[0])) for i in encoded]
        torch_inputs = torch.from_numpy(numpy.array(padded))
        yield from self._model.get_z_mean(torch_inputs).detach().cpu().numpy()

    @staticmethod
    def reduce_per_protein(embedding: ndarray) -> ndarray:
        return embedding



"""
Abstract interface for Embedder.

Authors:
  Christian Dallago
  Konstantin Schuetze
"""

import abc
import logging
import tempfile
from typing import List, Generator, Optional, Iterable, ClassVar, Any, Dict, Union

import torch
from numpy import ndarray

from bio_embeddings.utilities import (
    get_model_file,
    get_model_directories_from_zip,
    get_device,
)

logger = logging.getLogger(__name__)


class EmbedderInterface(abc.ABC):
    name: ClassVar[str]
    # An integer representing the size of the embedding.
    embedding_dimension: ClassVar[int]
    # An integer representing the number of layers from the RAW output of the LM.
    number_of_layers: ClassVar[int]
    # The files or directories with weights and config
    _necessary_files: ClassVar[List[str]] = []
    _necessary_directories: ClassVar[List[str]] = []
    _device: torch.device
    _options: Dict[str, Any]

    def __init__(self, device: Union[None, str, torch.device] = None, **kwargs):
        """
        Initializer accepts location of a pre-trained model and options
        """
        self._options = kwargs
        self._device = get_device(device)

        # Special case because SeqVec can currently be used with either a model directory or two files
        if self.__class__.__name__ == "SeqVecEmbedder":
            # No need to download weights_file/options_file if model_directory is given
            if "model_directory" in self._options:
                return

        files_loaded = 0
        for file in self._necessary_files:
            if not self._options.get(file):
                self._options[file] = get_model_file(model=self.name, file=file)
                files_loaded += 1

        for directory in self._necessary_directories:
            if not self._options.get(directory):
                self._options[directory] = get_model_directories_from_zip(
                    model=self.name, directory=directory
                )

                files_loaded += 1

        total_necessary = len(self._necessary_files) + len(self._necessary_directories)
        if 0 < files_loaded < total_necessary:
            logger.warning(
                f"You should pass either all necessary files or directories, or none, "
                f"while you provide {files_loaded} of {total_necessary}"
            )

    @abc.abstractmethod
    def embed(self, sequence: str) -> ndarray:
        """
        Returns embedding for one sequence.

        :param sequence: Valid amino acid sequence as String
        :return: An embedding of the sequence.
        """

        raise NotImplementedError

    def embed_batch(self, batch: List[str]) -> Generator[ndarray, None, None]:
        """Computes the embeddings from all sequences in the batch

        The provided implementation is dummy implementation that should be
        overwritten with the appropriate batching method for the model."""
        for sequence in batch:
            yield self.embed(sequence)

    def embed_many(
        self, sequences: Iterable[str], batch_size: Optional[int] = None
    ) -> Generator[ndarray, None, None]:
        """
        Returns embedding for one sequence.

        :param sequences: List of proteins as AA strings
        :param batch_size: For embedders that profit from batching, this is maximum number of AA per batch
        :return: A list object with embeddings of the sequences.
        """

        if batch_size:
            batch = []
            length = 0
            for sequence in sequences:
                if len(sequence) > batch_size:
                    logger.warning(
                        f"A sequence is {len(sequence)} residues long, "
                        f"which is longer than your `batch_size` parameter which is {batch_size}"
                    )
                    yield from self.embed_batch([sequence])
                    continue
                if length + len(sequence) >= batch_size:
                    yield from self.embed_batch(batch)
                    batch = []
                    length = 0
                batch.append(sequence)
                length += len(sequence)
            yield from self.embed_batch(batch)
        else:
            for seq in sequences:
                yield self.embed(seq)

    @staticmethod
    @abc.abstractmethod
    def reduce_per_protein(embedding: ndarray) -> ndarray:
        """
        For a variable size embedding, returns a fixed size embedding encoding all information of a sequence.

        :param embedding: the embedding
        :return: A fixed size embedding (a vector of size N, where N is fixed)
        """

        raise NotImplementedError


class EmbedderWithFallback(EmbedderInterface, abc.ABC):
    """ Batching embedder that will fallback to the CPU if the embedding on the GPU failed """

    _model: Any

    @abc.abstractmethod
    def _embed_batch_impl(
        self, batch: List[str], model: Any
    ) -> Generator[ndarray, None, None]:
        ...

    @abc.abstractmethod
    def _get_fallback_model(self):
        """ Returns a (cached) cpu model """
        ...

    def embed_batch(self, batch: List[str]) -> Generator[ndarray, None, None]:
        """Tries to get the embeddings in this order:
          * Full batch GPU
          * Single Sequence GPU
          * Single Sequence CPU

        Single sequence processing is done in case of runtime error due to
        a) very long sequence or b) too large batch size
        If this fails, you might want to consider lowering batch_size and/or
        cutting very long sequences into smaller chunks

        Returns unprocessed embeddings
        """
        # No point in having a fallback model when the normal model is CPU already
        if self._device.type == "cpu":
            yield from self._embed_batch_impl(batch, self._model)
            return

        try:
            yield from self._embed_batch_impl(batch, self._model)
        except RuntimeError as e:
            if len(batch) == 1:
                logger.error(
                    f"RuntimeError for sequence with {len(batch[0])} residues: {e}. "
                    f"This most likely means that you don't have enough GPU RAM to embed a protein this long. "
                    f"Embedding on the CPU instead, which is very slow"
                )
                yield from self._embed_batch_impl(batch, self._get_fallback_model())
            else:
                logger.error(
                    f"Error processing batch of {len(batch)} sequences: {e}. "
                    f"You might want to consider adjusting the `batch_size` parameter. "
                    f"Will try to embed each sequence in the set individually on the GPU."
                )
                for sequence in batch:
                    try:
                        yield from self._embed_batch_impl([sequence], self._model)
                    except RuntimeError as e:
                        logger.error(
                            f"RuntimeError for sequence with {len(sequence)} residues: {e}. "
                            f"This most likely means that you don't have enough GPU RAM to embed a protein this long."
                        )
                        yield from self._embed_batch_impl(
                            [sequence], self._get_fallback_model()
                        )



import logging
from pathlib import Path
from typing import List, Optional, Generator

from allennlp.commands.elmo import ElmoEmbedder
from numpy import ndarray

from bio_embeddings.embed.embedder_interfaces import EmbedderWithFallback

logger = logging.getLogger(__name__)

# A random short sequence (T0922 from CASP2)
_warmup_seq = "MGSSHHHHHHSSGLVPRGSHMASVQKFPGDANCDGIVDISDAVLIMQTMANPSKYQMTDKGRINADVTGNSDGVTVLDAQFIQSYCLGLVELPPVE"


class SeqVecEmbedder(EmbedderWithFallback):
    """SeqVec Embedder

    Heinzinger, Michael, et al. "Modeling aspects of the language of life
    through transfer-learning protein sequences." BMC bioinformatics 20.1 (2019): 723.
    https://doi.org/10.1186/s12859-019-3220-8
    """
    name = "seqvec"
    embedding_dimension = 1024
    number_of_layers = 3

    _weights_file: str
    _options_file: str
    _model: ElmoEmbedder
    # The fallback model running on the cpu, which will be initialized if needed
    _model_fallback: Optional[ElmoEmbedder] = None
    _necessary_files = ["weights_file", "options_file"]

    def __init__(self, warmup_rounds: int = 4, **kwargs):
        """
        Initialize Elmo embedder. Can define non-positional arguments for paths of files and other settings.

        :param warmup_rounds: A sample sequence will be embedded this often to
            work around elmo's non-determinism (https://github.com/allenai/allennlp/blob/v0.9.0/tutorials/how_to/elmo.md#notes-on-statefulness-and-non-determinism)
        :param weights_file: path of weights file
        :param options_file: path of options file
        :param model_directory: Alternative of weights_file/options_file
        :param max_amino_acids: max # of amino acids to include in embed_many batches. Default: 15k AA
        """
        super().__init__(**kwargs)

        # Get file locations from kwargs
        if "model_directory" in self._options:
            self._weights_file = str(
                Path(self._options["model_directory"]).joinpath("weights_file")
            )
            self._options_file = str(
                Path(self._options["model_directory"]).joinpath("options_file")
            )
        else:
            self._weights_file = self._options["weights_file"]
            self._options_file = self._options["options_file"]

        if self._device.type == "cuda":
            logger.info("CUDA available, using the GPU")
            cuda_device = self._device.index or 0
        else:
            logger.info("CUDA NOT available, using the CPU. This is slow")
            cuda_device = -1

        self._model = ElmoEmbedder(
            weight_file=self._weights_file,
            options_file=self._options_file,
            cuda_device=cuda_device,
        )

        self.warmup_rounds = warmup_rounds
        if self.warmup_rounds > 0:
            logger.info("Running ELMo warmup")
            for _ in range(self.warmup_rounds):
                self.embed(_warmup_seq)

    def embed(self, sequence: str) -> ndarray:
        return self._model.embed_sentence(list(sequence))

    def _get_fallback_model(self) -> ElmoEmbedder:
        if not self._model_fallback:
            logger.warning(
                "Loading model for CPU into RAM. Embedding on the CPU is very slow and you should avoid it."
            )
            self._model_fallback = ElmoEmbedder(
                weight_file=self._weights_file,
                options_file=self._options_file,
                cuda_device=-1,
            )
            if self.warmup_rounds > 0:
                logger.info("Running CPU ELMo warmup")
                for _ in range(self.warmup_rounds):
                    self._model_fallback.embed_sentence(list(_warmup_seq))
        return self._model_fallback

    def _embed_batch_impl(
        self, batch: List[str], model: ElmoEmbedder
    ) -> Generator[ndarray, None, None]:
        # elmo expect a `List[str]` as it was meant for tokens/words with more than one character.
        yield from model.embed_batch([list(seq) for seq in batch])

    @staticmethod
    def reduce_per_protein(embedding):
        return embedding.sum(0).mean(0)



"""Various helpers"""

from bio_embeddings.utilities.exceptions import *
from bio_embeddings.utilities.filemanagers import (
    get_file_manager,
    FileSystemFileManager,
)
from bio_embeddings.utilities.filemanagers.FileManagerInterface import (
    FileManagerInterface,
)
from bio_embeddings.utilities.helpers import *
from bio_embeddings.utilities.remote_file_retriever import (
    get_model_file,
    get_model_directories_from_zip,
)

__all__ = [
    "FileManagerInterface",
    "FileSystemFileManager",
    "InvalidParameterError",
    "check_required",
    "get_device",
    "get_file_manager",
    "get_model_directories_from_zip",
    "get_model_file",
    "read_fasta",
    "reindex_h5_file",
    "reindex_sequences",
    "write_fasta_file",
]



import collections
import h5py
import numpy as np

from typing import List
from sklearn.metrics import pairwise_distances as _pairwise_distances


PairwiseDistanceMatrixResult = collections.namedtuple('PairwiseDistanceMatrixResult', 'pairwise_matrix queries references')


def pairwise_distance_matrix_from_embeddings_and_annotations(query_embeddings_path: str, reference_embeddings_path: str,
                                                             metric: str = "euclidean", n_jobs: int = 1) -> PairwiseDistanceMatrixResult:
    """

    :param n_jobs: int, see scikit-learn documentation
    :param metric: Metric to use (string!), see scikit-learn documentation
    :param query_embeddings_path: A string defining a path to an h5 file
    :param reference_embeddings_path: A string defining a path to an h5 file
    :return: A tuple containing:
        - pairwise_matrix: the pairwise distances between queries and references
        - queries: A list of strings defining the queries
        - references: A list of strings defining the references
    """
    references: List[str]
    queries: List[str]
    reference_embeddings = list()
    query_embeddings = list()

    with h5py.File(reference_embeddings_path, 'r') as reference_embeddings_file,\
          h5py.File(query_embeddings_path, 'r') as query_embeddings_file:

        references = list(reference_embeddings_file.keys())
        queries = list(query_embeddings_file.keys())

        for refereince_identifier in references:
            reference_embeddings.append(np.array(reference_embeddings_file[refereince_identifier]))

        for query_identifier in queries:
            query_embeddings.append(np.array(query_embeddings_file[query_identifier]))

    pairwise_distances = _pairwise_distances(
        query_embeddings,
        reference_embeddings,
        metric=metric,
        n_jobs=n_jobs
    )

    return PairwiseDistanceMatrixResult(pairwise_matrix=pairwise_distances, queries=queries, references=references)


def get_k_nearest_neighbours(pairwise_matrix: np.array, k: int = 1) -> (List[int], np.array):
    """

    :param pairwise_matrix: an np.array with columns as queries and rows as targets
    :param k: the number of k-nn's to return
    :return: a list of tuples with indices of the nearest neighbour and distance to them (sorted by distance asc.)
    """
    resulting_indices = list()
    resulting_distances = list()

    for i, neighbour_distances in enumerate(pairwise_matrix):
        nearest_neighbour_indices = np.argpartition(neighbour_distances, k)[:k]
        nearest_neighbour_distances = np.array(list(map(neighbour_distances.__getitem__, nearest_neighbour_indices)))

        # nearest_neighbours will appear in an arbitrary order.
        # We want to ensure that the distances and indices are sorted by ascending distance
        # The following code shuffles both lists around to make sure that indices and distances are sorted equally
        nearest_neighbour_distances, nearest_neighbour_indices = (list(t) for t in zip(
            *sorted(
                zip(nearest_neighbour_distances, nearest_neighbour_indices)
            )
        ))

        resulting_indices.append(nearest_neighbour_indices)
        resulting_distances.append(nearest_neighbour_distances)

    return resulting_indices, np.array(resulting_distances)



from enum import Enum


class Location(Enum):

    CELL_MEMBRANE = 'Cell-Membrane'
    CYTOPLASM = 'Cytoplasm'
    ENDOPLASMATIC_RETICULUM = 'Endoplasmic reticulum'
    GOLGI_APPARATUS = 'Golgi - Apparatus'
    LYSOSOME_OR_VACUOLE = 'Lysosome / Vacuole'
    MITOCHONDRION = 'Mitochondrion'
    NUCLEUS = 'Nucleus'
    PEROXISOME = 'Peroxisome'
    PLASTID = 'Plastid'
    EXTRACELLULAR = 'Extra - cellular'
    UNKNOWN = '?'

    def __str__(self):
        return {
            self.CELL_MEMBRANE: "Cell-Membrane",
            self.CYTOPLASM: "Cytoplasm",
            self.ENDOPLASMATIC_RETICULUM: "Endoplasmic reticulum",
            self.GOLGI_APPARATUS: "Golgi - Apparatus",
            self.LYSOSOME_OR_VACUOLE: "Lysosome / Vacuole",
            self.MITOCHONDRION: 'Mitochondrion',
            self.NUCLEUS: 'Nucleus',
            self.PEROXISOME: 'Peroxisome',
            self.PLASTID: 'Plastid',
            self.EXTRACELLULAR: 'Extra - cellular',
            self.UNKNOWN: '?'
        }.get(self)

    @staticmethod
    def isAAFeature():
        return False



from pathlib import Path

from transformers import AlbertModel, AlbertTokenizer

from bio_embeddings.embed.prottrans_base_embedder import ProtTransBertBaseEmbedder


class ProtTransAlbertBFDEmbedder(ProtTransBertBaseEmbedder):
    """ProtTrans-Albert-BFD Embedder (ProtAlbert-BFD)

    Elnaggar, Ahmed, et al. "ProtTrans: Towards Cracking the Language of Life's
    Code Through Self-Supervised Deep Learning and High Performance Computing."
    arXiv preprint arXiv:2007.06225 (2020). https://arxiv.org/abs/2007.06225
    """
    _model: AlbertModel
    name = "prottrans_albert_bfd"
    embedding_dimension = 4096
    number_of_layers = 1

    def __init__(self, **kwargs):
        """
        Initialize Albert embedder.

        :param model_directory:
        """
        super().__init__(**kwargs)

        self._model_directory = self._options["model_directory"]

        # make model
        self._model = AlbertModel.from_pretrained(self._model_directory)
        self._model = self._model.eval().to(self._device)
        self._model_fallback = None
        self._tokenizer = AlbertTokenizer(
            str(Path(self._model_directory) / "albert_vocab_model.model"),
            do_lower_case=False,
        )

    def _get_fallback_model(self) -> AlbertModel:
        """ Returns the CPU model """
        if not self._model_fallback:
            self._model_fallback = AlbertModel.from_pretrained(
                self._model_directory
            ).eval()
        return self._model_fallback



from typing import Any, Dict, Union

import torch
from numpy import ndarray

from bio_embeddings.embed import EmbedderInterface


class UniRepEmbedder(EmbedderInterface):
    """UniRep Embedder

    Alley, E.C., Khimulya, G., Biswas, S. et al. Unified rational protein
    engineering with sequence-based deep representation learning. Nat Methods
    16, 1315–1322 (2019). https://doi.org/10.1038/s41592-019-0598-1

    We use a reimplementation of unirep:

    Ma, Eric, and Arkadij Kummer. "Reimplementing Unirep in JAX." bioRxiv (2020).
    https://doi.org/10.1101/2020.05.11.088344
    """

    name = "unirep"
    # An integer representing the size of the embedding.
    embedding_dimension = 1900
    # An integer representing the number of layers from the RAW output of the LM.
    number_of_layers = 1

    params: Dict[str, Any]

    def __init__(self, device: Union[None, str, torch.device] = None, **kwargs):
        from jax_unirep.utils import load_params_1900

        if device:
            raise NotImplementedError("UniRep does not allow configuring the device")
        super().__init__(device, **kwargs)

        self.params = load_params_1900()

    def embed(self, sequence: str) -> ndarray:
        from jax import vmap, partial
        from jax_unirep.featurize import apply_fun
        from jax_unirep.utils import get_embeddings

        # Unirep only allows batching with sequences of the same length, so we don't do batching at all
        embedded_seqs = get_embeddings([sequence])
        # h and c refer to hidden and cell state
        # h contains all the hidden states, while h_final and c_final contain only the last state
        h_final, c_final, h = vmap(partial(apply_fun, self.params))(embedded_seqs)
        # Go from a batch of 1, which is `(1, len(sequence), 1900)`, to `len(sequence), 1900)`
        return h[0]

    @staticmethod
    def reduce_per_protein(embedding: ndarray) -> ndarray:
        # This is `h_avg` in jax-unirep terminology
        return embedding.mean(axis=0)



import h5py
import numpy as np
from copy import deepcopy
from pandas import read_csv
from bio_embeddings.utilities import InvalidParameterError, check_required, get_file_manager
from bio_embeddings.project.tsne import tsne_reduce
from bio_embeddings.project.umap import umap_reduce


def tsne(**kwargs):
    result_kwargs = deepcopy(kwargs)
    file_manager = get_file_manager(**kwargs)

    # Get sequence mapping to use as information source
    mapping = read_csv(result_kwargs['mapping_file'], index_col=0)

    reduced_embeddings_file_path = result_kwargs['reduced_embeddings_file']

    reduced_embeddings = []

    with h5py.File(reduced_embeddings_file_path, 'r') as f:
        for remapped_id in mapping.index:
            reduced_embeddings.append(np.array(f[str(remapped_id)]))

    # Get parameters or set defaults
    result_kwargs['perplexity'] = kwargs.get('perplexity', 6)
    result_kwargs['n_jobs'] = kwargs.get('n_jobs', -1)
    result_kwargs['n_iter'] = kwargs.get('n_iter', 15000)

    result_kwargs['metric'] = kwargs.get('metric', 'cosine')
    result_kwargs['n_components'] = kwargs.get('n_components', 3)
    result_kwargs['random_state'] = kwargs.get('random_state', 420)
    result_kwargs['verbose'] = kwargs.get('verbose', 1)

    projected_embeddings = tsne_reduce(reduced_embeddings, **kwargs)

    for i in range(result_kwargs['n_components']):
        mapping[f'component_{i}'] = projected_embeddings[:, i]

    projected_embeddings_file_path = file_manager.create_file(kwargs.get('prefix'),
                                                              result_kwargs.get('stage_name'),
                                                              'projected_embeddings_file',
                                                              extension='.csv')

    mapping.to_csv(projected_embeddings_file_path)
    result_kwargs['projected_embeddings_file'] = projected_embeddings_file_path

    return result_kwargs


def umap(**kwargs):
    result_kwargs = deepcopy(kwargs)
    file_manager = get_file_manager(**kwargs)

    # Get sequence mapping to use as information source
    mapping = read_csv(result_kwargs['mapping_file'], index_col=0)

    reduced_embeddings_file_path = result_kwargs['reduced_embeddings_file']

    reduced_embeddings = []

    with h5py.File(reduced_embeddings_file_path, 'r') as f:
        for remapped_id in mapping.index:
            reduced_embeddings.append(np.array(f[str(remapped_id)]))

    # Get parameters or set defaults
    result_kwargs['min_dist'] = kwargs.get('min_dist', .6)
    result_kwargs['n_neighbors'] = kwargs.get('n_neighbors', 15)
    result_kwargs['spread'] = kwargs.get('spread', 1)
    result_kwargs['metric'] = kwargs.get('metric', 'cosine')
    result_kwargs['n_components'] = kwargs.get('n_components', 3)
    result_kwargs['random_state'] = kwargs.get('random_state', 420)
    result_kwargs['verbose'] = kwargs.get('verbose', 1)

    projected_embeddings = umap_reduce(reduced_embeddings, **kwargs)

    for i in range(result_kwargs['n_components']):
        mapping[f'component_{i}'] = projected_embeddings[:, i]

    projected_embeddings_file_path = file_manager.create_file(kwargs.get('prefix'),
                                                              result_kwargs.get('stage_name'),
                                                              'projected_embeddings_file',
                                                              extension='.csv')

    mapping.to_csv(projected_embeddings_file_path)
    result_kwargs['projected_embeddings_file'] = projected_embeddings_file_path

    return result_kwargs


# list of available projection protocols
PROTOCOLS = {
    "tsne": tsne,
    "umap": umap,
}


def run(**kwargs):
    """
    Run project protocol

    Parameters
    ----------
    kwargs arguments (* denotes optional):
        reduced_embeddings_file: Where per-protein embeddings live
        prefix: Output prefix for all generated files
        stage_name: The stage name
        protocol: Which projection technique to use
        mapping_file: the mapping file generated by the pipeline when remapping indexes

    Returns
    -------
    Dictionary with results of stage
    """
    check_required(kwargs, ['protocol', 'prefix', 'stage_name', 'reduced_embeddings_file', 'mapping_file'])

    if kwargs["protocol"] not in PROTOCOLS:
        raise InvalidParameterError(
            "Invalid protocol selection: " +
            "{}. Valid protocols are: {}".format(
                kwargs["protocol"], ", ".join(PROTOCOLS.keys())
            )
        )

    return PROTOCOLS[kwargs["protocol"]](**kwargs)


from enum import Enum


class Membrane(Enum):

    MEMBRANE = 'Membrane bound'
    SOLUBLE = 'Soluble'
    UNKNOWN = "?"

    def __str__(self):
        return {
            self.MEMBRANE: "Membrane bound",
            self.SOLUBLE: "Soluble",
            self.UNKNOWN: "Unknown"
        }.get(self)

    @staticmethod
    def isAAFeature():
        return False



import logging
import torch
import collections

from typing import List, Union
from numpy import ndarray
from enum import Enum

from bio_embeddings.extract.annotations import Location, Membrane, Disorder, SecondaryStructure
from bio_embeddings.extract.basic.annotation_inference_models import SUBCELL_FNN, SECSTRUCT_CNN
from bio_embeddings.utilities import get_device, get_model_file

logger = logging.getLogger(__name__)

# Label mappings
_loc_labels = {
    0: Location.CELL_MEMBRANE,
    1: Location.CYTOPLASM,
    2: Location.ENDOPLASMATIC_RETICULUM,
    3: Location.GOLGI_APPARATUS,
    4: Location.LYSOSOME_OR_VACUOLE,
    5: Location.MITOCHONDRION,
    6: Location.NUCLEUS,
    7: Location.PEROXISOME,
    8: Location.PLASTID,
    9: Location.EXTRACELLULAR
}

_mem_labels = {
    0: Membrane.SOLUBLE,
    1: Membrane.MEMBRANE
}

_dssp8_labels = {
    0: SecondaryStructure.THREE_HELIX,
    1: SecondaryStructure.ALPHA_HELIX,
    2: SecondaryStructure.FIVE_HELIX,
    3: SecondaryStructure.ISOLATED_BETA_BRIDGE,
    4: SecondaryStructure.EXTENDED_STRAND,
    5: SecondaryStructure.BEND,
    6: SecondaryStructure.TURN,
    7: SecondaryStructure.IRREGULAR
}

_dssp3_labels = {
    0: SecondaryStructure.ALPHA_HELIX,
    1: SecondaryStructure.EXTENDED_STRAND,
    2: SecondaryStructure.IRREGULAR
}

_disor_labels = {
    0: Disorder.ORDER,
    1: Disorder.DISORDER
}

BasicSecondaryStructureResult = collections.namedtuple('BasicSecondaryStructureResult', 'DSSP3 DSSP8 disorder')
BasicSubcellularLocalizationResult = collections.namedtuple('BasicSubcellularLocalizationResult', 'localization membrane')
BasicExtractedAnnotations = collections.namedtuple('BasicExtractedAnnotations', 'DSSP3 DSSP8 disorder localization membrane')


class BasicAnnotationExtractor(object):
    necessary_files = ["secondary_structure_checkpoint_file", "subcellular_location_checkpoint_file"]

    def __init__(self, model_type: str, device: Union[None, str, torch.device] = None, **kwargs):
        """
        Initialize annotation extractor. Must define non-positional arguments for paths of files.

        :param secondary_structure_checkpoint_file: path of secondary structure inference model checkpoint file
        :param subcellular_location_checkpoint_file: path of the subcellular location inference model checkpoint file
        """

        self._options = kwargs
        self._model_type = model_type
        self._device = get_device(device)

        # Create un-trained (raw) model and ensure self._model_type is valid
        if self._model_type == "seqvec_from_publication":
            self._subcellular_location_model = SUBCELL_FNN().to(self._device)
        elif self._model_type == "bert_from_publication": # Drop batchNorm for ProtTrans models
            self._subcellular_location_model = SUBCELL_FNN(use_batch_norm=False).to(self._device)
        else:
            print("You first need to define your custom model architecture.")
            raise NotImplementedError

        # Download the checkpoint files if needed
        for file in self.necessary_files:
            if not self._options.get(file):
                self._options[file] = get_model_file(model=f"{self._model_type}_annotations_extractors", file=file)

        self._secondary_structure_checkpoint_file = self._options['secondary_structure_checkpoint_file']
        self._subcellular_location_checkpoint_file = self._options['subcellular_location_checkpoint_file']

        # Read in pre-trained model

        self._secondary_structure_model = SECSTRUCT_CNN().to(self._device)

        # load pre-trained weights for annotation machines
        subcellular_state = torch.load(self._subcellular_location_checkpoint_file, map_location=self._device)
        secondary_structure_state = torch.load(self._secondary_structure_checkpoint_file, map_location=self._device)

        # load pre-trained weights into raw model
        self._subcellular_location_model.load_state_dict(subcellular_state['state_dict'])
        self._secondary_structure_model.load_state_dict(secondary_structure_state['state_dict'])

        # ensure that model is in evaluation mode (important for batchnorm and dropout)
        self._subcellular_location_model.eval()
        self._secondary_structure_model.eval()

    def get_subcellular_location(self, raw_embedding: ndarray) -> BasicSubcellularLocalizationResult:
        # Reduce embedding to fixed size, per-sequence (aka: Lx3x2014 --> 1024).
        # This is similar to embedder.reduce_per_protein(),
        # but more efficient since may be run in GPU (see self._device)

        # TODO: xxmh I forgot that SeqVec requires different pooling to derive fixed size rep.
        #   SeqVec requires summing over 3 layers, ProtTrans models only extract last layers
        #   Quick&Dirty solution is to check for shape of embedding tensors as SeqVec has 3 dims,
        #   while ProtTrans should only have 2 dims.
        #   Better way would be to access some internal variable (probably I just missed this flag)
        #   XXCD: can check embedder type via protol in embed config, but this may become complicated...
        if self._model_type == "seqvec_from_publication":
            # SeqVec case
            embedding = torch.tensor(raw_embedding).to(self._device).sum(dim=0).mean(dim=0, keepdim=True)
        elif self._model_type == "bert_from_publication":
            # Bert case
            embedding = torch.tensor(raw_embedding).to(self._device).mean(dim=0, keepdim=True)
        else:
            raise NotImplementedError

        yhat_loc, yhat_mem = self._subcellular_location_model(embedding)

        pred_loc = _loc_labels[torch.max(yhat_loc, dim=1)[1].item()]  # get index of output node with max. activation,
        pred_mem = _mem_labels[torch.max(yhat_mem, dim=1)[1].item()]  # this corresponds to the predicted class

        return BasicSubcellularLocalizationResult(localization=pred_loc, membrane=pred_mem)

    def get_secondary_structure(self, raw_embedding: ndarray) -> BasicSecondaryStructureResult:
        # TODO: xxmh: same as for subcell loc.:
        #       SeqVec requires summing over layers while ProtTrans models only extract last layers
        if self._model_type == "seqvec_from_publication":
            # SeqVec case
            embedding = torch.tensor(raw_embedding).to(self._device).sum(dim=0, keepdim=True).permute(0, 2, 1).unsqueeze(dim=-1)
        elif self._model_type == "bert_from_publication":
            # Bert case
            # Flip dimensions for ProtTrans models in order to make feature dimension the first dimension
            embedding = torch.tensor(raw_embedding).to(self._device).T[None, :, :, None]
        else:
            raise NotImplementedError

        yhat_dssp3, yhat_dssp8, yhat_disor = self._secondary_structure_model(embedding)

        pred_dssp3 = self._class2label(_dssp3_labels, yhat_dssp3)
        pred_dssp8 = self._class2label(_dssp8_labels, yhat_dssp8)
        pred_disor = self._class2label(_disor_labels, yhat_disor)

        return BasicSecondaryStructureResult(DSSP3=pred_dssp3, DSSP8=pred_dssp8, disorder=pred_disor)

    def get_annotations(self, raw_embedding: ndarray) -> BasicExtractedAnnotations:
        secstruct = self.get_secondary_structure(raw_embedding)
        subcell = self.get_subcellular_location(raw_embedding)

        return BasicExtractedAnnotations(disorder=secstruct.disorder, DSSP8=secstruct.DSSP8,
                                         DSSP3=secstruct.DSSP3, localization=subcell.localization,
                                         membrane=subcell.membrane)

    @staticmethod
    def _class2label(label_dict, yhat) -> List[Enum]:
        # get index of output node with max. activation (=predicted class)
        class_indices = torch.max(yhat, dim=1)[1].squeeze()
        return [label_dict[class_idx.item()] for class_idx in class_indices]



#!/usr/bin/env python3

import argparse
import logging

from bio_embeddings.utilities.pipeline import parse_config_file_and_execute_run


def main():
    """
    Pipeline commandline entry point
    """
    logging.basicConfig(level=logging.INFO, format="%(asctime)s %(levelname)s %(message)s")
    # Jax likes to print warnings
    logging.captureWarnings(True)

    parser = argparse.ArgumentParser(description='Embeds protein sequences.')
    parser.add_argument('-o', '--overwrite', dest='overwrite', required=False, action='store_true',
                        help='Will force overwrite of previously existing results.')
    parser.add_argument('config_path', metavar='/path/to/pipeline_definition.yml', type=str, nargs=1,
                        help='The path to the config. For examples, see folder "parameter examples".')
    arguments = parser.parse_args()

    parse_config_file_and_execute_run(arguments.config_path[0], overwrite=arguments.overwrite)


if __name__ == '__main__':
    main()



from enum import Enum


class SecondaryStructure(Enum):
    # From https://swift.cmbi.umcn.nl/gv/dssp/

    ALPHA_HELIX = "H"
    ISOLATED_BETA_BRIDGE = "B"
    EXTENDED_STRAND = "E"
    THREE_HELIX = "G"
    FIVE_HELIX = "I"
    TURN = "T"
    BEND = "S"
    IRREGULAR = "C"
    UNKNOWN = "?"

    def __str__(self):
        return {
            self.ALPHA_HELIX: "α-helix",
            self.ISOLATED_BETA_BRIDGE: "Residue in isolated β-bridge",
            self.EXTENDED_STRAND: "Extended strand, participates in β ladder",
            self.THREE_HELIX: "3-helix",
            self.FIVE_HELIX: "5 helix",
            self.TURN: "Hydrogen bonded turn",
            self.BEND: "Bend",
            self.IRREGULAR: "Loop/Irregular",
            self.UNKNOWN: "Unknown"
        }.get(self)

    @staticmethod
    def isAAFeature():
        return True



#!/usr/bin/env python3
"""
Figure out what is the longest sequences that can be embedded on the GPU
"""
import logging
from argparse import ArgumentParser
from pathlib import Path
from typing import Optional

from bio_embeddings.embed import name_to_embedder

logger = logging.getLogger(__name__)


def bisect_embedder_memory(embedder_name: str, model_directory: Optional[str]):
    logging.basicConfig(level=logging.INFO, format="%(asctime)s %(message)s")
    fasta_lines = Path("test-data/titin_twice.fasta").read_text().splitlines()
    titin = "".join(fasta_lines[1:]).replace("\n", "")
    embedder = name_to_embedder[embedder_name]
    if model_directory:
        embedder = embedder(
            model_directory=str(Path(model_directory).joinpath(embedder.name))
        )
    else:
        logger.info("Model directory not given, downloading model")
        embedder = embedder()
    # Evil Hack: `not model_fallback` will now evaluate to False, but using it will still fail
    embedder.model_fallback = True

    # Binary search
    low = 0
    high = len(titin)
    while True:
        middle = (low + high) // 2
        try:
            list(embedder.embed_many([titin[:middle]], batch_size=100000))
            low = middle
            print(f"{middle} passed, new range {low} to {high}")
        except Exception as e:
            high = middle
            print(
                f"{middle} failed, new range {low} to {high}: {type(e).__name__}: {e}"
            )
        if high - low == 1:
            print(f"Max amino acids for {embedder_name}: {low}")
            break


def main():
    parser = ArgumentParser()
    parser.add_argument("embedder", help="Lowercase name of the embedder")
    parser.add_argument("--model-directory")
    args = parser.parse_args()
    if args.embedder == "all":
        for embedder_name in name_to_embedder.keys():
            bisect_embedder_memory(
                embedder_name, str(Path(args.model_directory).joinpath(embedder_name))
            )
    else:
        bisect_embedder_memory(args.embedder, args.model_directory)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bio_embeddings',
 'bio_embeddings.embed',
 'bio_embeddings.extract',
 'bio_embeddings.extract.annotations',
 'bio_embeddings.extract.basic',
 'bio_embeddings.project',
 'bio_embeddings.utilities',
 'bio_embeddings.utilities.filemanagers',
 'bio_embeddings.visualize']

package_data = \
{'': ['*']}

install_requires = \
['appdirs>=1.4.4,<2.0.0',
 'biopython>=1.76,<2.0',
 'gensim>=3.8.2,<4.0.0',
 'h5py>=2.10.0,<3.0.0',
 'importlib_metadata>=1.7.0,<2.0.0',
 'jaxlib==0.1.57',
 'lock>=2018.3.25,<2019.0.0',
 'matplotlib>=3.2.1,<4.0.0',
 'numpy>=1.18.3,<2.0.0',
 'pandas>=1.0.3,<2.0.0',
 'plotly>=4.6.0,<5.0.0',
 'ruamel.yaml>=0.16.10,<0.17.0',
 'scikit-learn>=0.22.2.post1,<0.23.0',
 'scipy>=1.4.1,<2.0.0',
 'torch>=1.5.0,<1.6.0',
 'tqdm>=4.45.0,<5.0.0',
 'umap-learn>=0.4.2,<0.5.0',
 'urllib3==1.25.10']

extras_require = \
{'all': ['bio-embeddings-esm==0.2.1',
         'bio-embeddings-cpcprot==0.0.1',
         'bio-embeddings-tape-proteins==0.4',
         'bio-embeddings-plus==0.1.0',
         'bio-embeddings-bepler==0.0.1',
         'bio-embeddings-allennlp==0.9.0',
         'transformers>=3.1.0,<4.0.0',
         'jax-unirep>=1.0.1,<2.0.0'],
 'bepler': ['bio-embeddings-bepler==0.0.1'],
 'cpcprot': ['bio-embeddings-cpcprot==0.0.1',
             'bio-embeddings-tape-proteins==0.4'],
 'esm': ['bio-embeddings-esm==0.2.1'],
 'plus': ['bio-embeddings-plus==0.1.0'],
 'seqvec': ['bio-embeddings-allennlp==0.9.0',
            'boto3==1.14.18',
            'botocore==1.17.18'],
 'transformers': ['transformers>=3.1.0,<4.0.0'],
 'unirep': ['jax-unirep>=1.0.1,<2.0.0'],
 'webserver': ['pymongo>=3.11.2,<4.0.0', 'sentry-sdk[flask]>=0.19.5,<0.20.0']}

entry_points = \
{'console_scripts': ['bio_embeddings = bio_embeddings.utilities.cli:main']}

setup_kwargs = {
    'name': 'bio-embeddings',
    'version': '0.1.5b1',
    'description': 'A pipeline for protein embedding generation and visualization',
    'long_description': '<p align="center">\n  <a href="https://chat.bioembeddings.com/">\n    <img src="https://chat.bioembeddings.com/api/v1/shield.svg?type=online&name=chat&icon=false" />\n  </a>\n</p>\n\n# Bio Embeddings\nResources to learn about bio_embeddings:\n\n- Quickly predict protein structure and function from sequence via embeddings: [embed.protein.properties](https://embed.protein.properties).\n- Read the current documentation: [docs.bioembeddings.com](https://docs.bioembeddings.com).\n- Chat with us: [chat.bioembeddings.com](https://chat.bioembeddings.com).\n- We presented the bio_embeddings pipeline as a talk at ISMB 2020 & LMRL 2020. You can [find the talk on YouTube](https://www.youtube.com/watch?v=NucUA0QiOe0&feature=youtu.be), and [the poster on F1000](https://f1000research.com/posters/9-876).\n- Check out the [`examples`](examples) of pipeline configurations a and [`notebooks`](notebooks).\n\nProject aims:\n\n  - Facilitate the use of language model based biological sequence representations for transfer-learning by providing a single, consistent interface and close-to-zero-friction\n  - Reproducible workflows\n  - Depth of representation (different models from different labs trained on different dataset for different purposes)\n  - Extensive examples, handle complexity for users (e.g. CUDA OOM abstraction) and well documented warnings and error messages.\n\nThe project includes:\n\n- General purpose python embedders based on open models trained on biological sequence representations (SeqVec, ProtTrans, UniRep,...)\n- A pipeline which:\n  - embeds sequences into matrix-representations (per-amino-acid) or vector-representations (per-sequence) that can be used to train learning models or for analytical purposes\n  - projects per-sequence embedidngs into lower dimensional representations using UMAP or t-SNE (for lightwieght data handling and visualizations)\n  - visualizes low dimensional sets of per-sequence embeddings onto 2D and 3D interactive plots (with and without annotations)\n  - extracts annotations from per-sequence and per-amino-acid embeddings using supervised (when available) and unsupervised approaches (e.g. by network analysis)\n- A webserver that wraps the pipeline into a distributed API for scalable and consistent workfolws\n\n## Installation\n\nYou can install `bio_embeddings` via pip or use it via docker.\n\n### Pip\n\nInstall the pipeline like so:\n\n```bash\npip install bio-embeddings[all]\n```\n\nTo get the latest features, please install the pipeline like so:\n\n```bash\npip install -U "bio-embeddings[all] @ git+https://github.com/sacdallago/bio_embeddings.git"\n```\n\n### Docker\n\nWe provide a docker image at `rostlab/bio_embeddings`. Simple usage example:\n\n```shell_script\ndocker run --rm --gpus all \\\n    -v "$(pwd)/examples/docker":/mnt \\\n    -u $(id -u ${USER}):$(id -g ${USER}) \\\n    rostlab/bio_embeddings /mnt/config.yml\n```\n\nSee the [`docker`](examples/docker) example in the [`examples`](examples) folder for instructions. We currently have published `rostlab/bio_embeddings:develop`. For our next stable release, we will publish tags for all releases and a `latest` tag pointing to the latest release.\n\n### Installation notes:\n\n`bio_embeddings` was developed for unix machines with GPU capabilities and [CUDA](https://developer.nvidia.com/cuda-zone) installed. If your setup diverges from this, you may encounter some inconsitencies (e.g. speed is significantly affected by the absence of a GPU and CUDA). For Windows users, we strongly recommend the use of [Windows Subsystem for Linux](https://docs.microsoft.com/en-us/windows/wsl/install-win10).\n\n\n## What model is right for you?\n\nEach models has its strengths and weaknesses (speed, specificity, memory footprint...). There isn\'t a "one-fits-all" and we encourage you to at least try two different models when attempting a new exploratory project.\n\nThe models `prottrans_bert_bfd`, `prottrans_albert_bfd`, `seqvec` and `prottrans_xlnet_uniref100` were all trained with the goal of systematic predictions. From this pool, we believe the optimal model to be `prottrans_bert_bfd`, followed by `seqvec`, which has been established for longer and uses a different principle (LSTM vs Transformer).\n\n## Usage and examples\n\nWe highly recommend you to check out the [`examples`](examples) folder for pipeline examples, and the [`notebooks`](notebooks) folder for post-processing pipeline runs and general purpose use of the embedders.\n\nAfter having installed the package, you can:\n\n1. Use the pipeline like:\n\n    ```bash\n    bio_embeddings config.yml\n    ```\n\n    [A blueprint of the configuration file](examples/parameters_blueprint.yml), and an example setup can be found in the [`examples`](examples) directory of this repository.\n\n1. Use the general purpose embedder objects via python, e.g.:\n\n    ```python\n    from bio_embeddings.embed import SeqVecEmbedder\n\n    embedder = SeqVecEmbedder()\n\n    embedding = embedder.embed("SEQVENCE")\n    ```\n\n    More examples can be found in the [`notebooks`](notebooks) folder of this repository.\n    \n## Cite\n\nWhile we are working on a proper publication, if you are already using this tool, we would appreciate if you could cite the following poster:\n\n> Dallago C, Schütze K, Heinzinger M et al. bio_embeddings: python pipeline for fast visualization of protein features extracted by language models [version 1; not peer reviewed]. F1000Research 2020, 9(ISCB Comm J):876 (poster) (doi: [10.7490/f1000research.1118163.1](https://doi.org/10.7490/f1000research.1118163.1))\n\n## Contributors\n\n- Christian Dallago (lead)\n- Konstantin Schütze\n- Tobias Olenyi\n- Michael Heinzinger\n\n----\n\n## Development status\n\n\n<details>\n<summary>Pipeline stages</summary>\n<br>\n\n- embed:\n  - [x] ProtTrans BERT trained on BFD (https://doi.org/10.1101/2020.07.12.199554)\n  - [x] SeqVec (https://bmcbioinformatics.biomedcentral.com/articles/10.1186/s12859-019-3220-8)\n  - [x] ProtTrans ALBERT trained on BFD (https://doi.org/10.1101/2020.07.12.199554)\n  - [x] ProtTrans XLNet trained on UniRef100 (https://doi.org/10.1101/2020.07.12.199554)\n  - [ ] Fastext\n  - [ ] Glove\n  - [ ] Word2Vec\n  - [x] UniRep (https://www.nature.com/articles/s41592-019-0598-1)\n  - [x] ESM (https://www.biorxiv.org/content/10.1101/622803v3)\n  - [x] PLUS (https://github.com/mswzeus/PLUS/)\n  - [x] CPCProt (https://www.biorxiv.org/content/10.1101/2020.09.04.283929v1.full.pdf)\n- project:\n  - [x] t-SNE\n  - [x] UMAP\n- visualize:\n  - [x] 2D/3D sequence embedding space\n- extract:\n  - supervised:\n    - [x] SeqVec: DSSP3, DSSP8, disorder, subcellular location and membrane boundness as in https://bmcbioinformatics.biomedcentral.com/articles/10.1186/s12859-019-3220-8\n    - [x] Bert: DSSP3, DSSP8, disorder, subcellular location and membrane boundness as in https://doi.org/10.1101/2020.07.12.199554\n  - unsupervised:\n    - [x] via sequence-level (reduced_embeddings), pairwise distance (euclidean like [goPredSim](https://github.com/Rostlab/goPredSim), more options available, e.g. cosine)\n</details>\n\n<details>\n<summary>Web server (unpublished)</summary>\n<br>\n\n- [x] SeqVec supervised predictions\n- [x] Bert supervised predictions\n- [ ] SeqVec unsupervised predictions for GO: CC, BP,..\n- [ ] Bert unsupervised predictions for GO: CC, BP,..\n- [ ] SeqVec unsupervised predictions for SwissProt (just a link to the 1st-k-nn)\n- [ ] Bert unsupervised predictions for SwissProt (just a link to the 1st-k-nn)\n</details>\n\n<details>\n<summary>General purpose embedders</summary>\n<br>\n\n- [x] ProtTrans BERT trained on BFD (https://doi.org/10.1101/2020.07.12.199554)\n- [x] SeqVec (https://bmcbioinformatics.biomedcentral.com/articles/10.1186/s12859-019-3220-8)\n- [x] ProtTrans ALBERT trained on BFD (https://doi.org/10.1101/2020.07.12.199554)\n- [x] ProtTrans XLNet trained on UniRef100 (https://doi.org/10.1101/2020.07.12.199554)\n- [x] Fastext\n- [x] Glove\n- [x] Word2Vec\n- [x] UniRep (https://www.nature.com/articles/s41592-019-0598-1)\n- [x] ESM (https://www.biorxiv.org/content/10.1101/622803v3)\n- [x] PLUS (https://github.com/mswzeus/PLUS/)\n- [x] CPCProt (https://www.biorxiv.org/content/10.1101/2020.09.04.283929v1.full.pdf)\n</details>\n',
    'author': 'Christian Dallago',
    'author_email': 'christian.dallago@tum.de',
    'maintainer': 'Rostlab',
    'maintainer_email': 'admin@rostlab.org',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6.6,<4.0',
}


setup(**setup_kwargs)



from bio_embeddings.utilities.filemanagers.FileSystemFileManager import FileSystemFileManager
from bio_embeddings.utilities.filemanagers.FileManagerInterface import FileManagerInterface


def get_file_manager(**kwargs):

    # A useless call to pacify the linters
    # TODO: when new FileManagers are available, parse the file manager type from "management".
    kwargs.get('management', {})

    return FileSystemFileManager()



from bio_embeddings.extract.annotations.Disorder import Disorder
from bio_embeddings.extract.annotations.Location import Location
from bio_embeddings.extract.annotations.Membrane import Membrane
from bio_embeddings.extract.annotations.SecondaryStructure import SecondaryStructure

__all__ = ["Disorder", "Location", "Membrane", "SecondaryStructure"]



from typing import Union, List, Generator

import torch
from numpy import ndarray
from plus.config import ModelConfig, RunConfig
from plus.data.alphabets import Protein
from plus.data.dataset import Embedding_dataset, collate_sequences_for_embedding
from plus.model.plus_rnn import PLUS_RNN, get_embedding
from plus.train import Trainer
from plus.utils import set_seeds
from torch.utils.data import DataLoader

from bio_embeddings.embed import EmbedderInterface


class PLUSRNNEmbedder(EmbedderInterface):
    """PLUS RNN Embedder

    Pre-Training of Deep Bidirectional Protein Sequence Representations with Structural Information
    Seonwoo Min, Seunghyun Park, Siwon Kim, Hyun-Soo Choi, Sungroh Yoon
    https://arxiv.org/abs/1912.05625"""

    name = "plus_rnn"
    number_of_layers = 1
    embedding_dimension = 1024

    _necessary_files = ["model_file"]

    _alphabet: Protein
    _model: PLUS_RNN
    _model_cfg: ModelConfig
    _run_cfg: RunConfig

    def __init__(self, device: Union[None, str, torch.device] = None, **kwargs):
        super().__init__(device, **kwargs)
        # This seed is copied from PLUS
        set_seeds(2020)

        # We inlined the config json files since they aren't shipped with the package
        self._alphabet = Protein()
        self._model_cfg = ModelConfig(input_dim=len(self._alphabet))
        self._model_cfg.model_type = "RNN"
        self._model_cfg.rnn_type = "B"
        self._model_cfg.num_layers = 3
        self._model_cfg.hidden_dim = 512
        self._model_cfg.embedding_dim = 100
        self._run_cfg = RunConfig(sanity_check=True)
        self._run_cfg.batch_size_eval = 512
        self._model = PLUS_RNN(self._model_cfg)
        self._model.load_weights(self._options["model_file"])
        self._model = self._model.to(self._device)

    def embed_batch(self, batch: List[str]) -> Generator[ndarray, None, None]:
        sequences = [
            self._alphabet.encode(sequence.encode().upper()) for sequence in batch
        ]
        test_dataset = [torch.from_numpy(sequence).long() for sequence in sequences]
        test_dataset = Embedding_dataset(
            test_dataset, self._alphabet, self._run_cfg, True
        )

        iterator_test = DataLoader(
            test_dataset,
            self._run_cfg.batch_size_eval,
            collate_fn=collate_sequences_for_embedding,
        )

        model_list = [self._model, "", True, False, False]
        tasks_list = [["", [], []]]  # list of lists [idx, metrics_train, metrics_eval]
        trainer = Trainer([model_list], get_embedding, self._run_cfg, tasks_list)
        for tokens, lengths in iterator_test:
            # https://github.com/pytorch/pytorch/issues/43227
            batch = (tokens.to(self._device), lengths)
            trainer.embed(batch, {"data_parallel": False})

        embeddings = trainer.tasks_dict["results_eval"][0]["embeddings"]
        # 1 is d_h with 1024 dimensions
        for i in range(len(embeddings[0])):
            yield embeddings[1][i].numpy()

        trainer.reset()

    def embed(self, sequence: str) -> ndarray:
        [embedding] = self.embed_batch([sequence])
        return embedding

    @staticmethod
    def reduce_per_protein(embedding: ndarray) -> ndarray:
        return embedding.mean(axis=0)



import logging
import re
from itertools import zip_longest
from typing import List, Generator, TypeVar, Union

import torch
from numpy import ndarray
from transformers import BertTokenizer, AlbertTokenizer, BertModel, AlbertModel

from bio_embeddings.embed.embedder_interfaces import EmbedderWithFallback

# https://stackoverflow.com/a/39205612/3549270
RealProtTransBertEmbedder = TypeVar(
    "RealProtTransEmbedder", bound="RealProtTransEmbedder"
)

logger = logging.getLogger(__name__)


class ProtTransBertBaseEmbedder(EmbedderWithFallback):
    """ Shared code between the ProtTrans models Bert and Albert """

    _tokenizer: Union[AlbertTokenizer, BertTokenizer]
    _model: Union[AlbertModel, BertModel]

    _necessary_directories = ["model_directory"]

    def _get_fallback_model(self) -> Union[BertModel, AlbertModel]:
        raise NotImplementedError

    def _embed_batch_impl(
        self, batch: List[str], model: Union[BertModel, AlbertModel]
    ) -> Generator[ndarray, None, None]:
        """ Embed batch code shared between Bert and Albert """
        seq_lens = [len(seq) for seq in batch]
        # Remove rare amino acids
        batch = [re.sub(r"[UZOB]", "X", sequence) for sequence in batch]
        # transformers needs spaces between the amino acids
        batch = [" ".join(list(seq)) for seq in batch]

        ids = self._tokenizer.batch_encode_plus(
            batch, add_special_tokens=True, pad_to_max_length=True
        )

        tokenized_sequences = torch.tensor(ids["input_ids"]).to(model.device)
        attention_mask = torch.tensor(ids["attention_mask"]).to(model.device)

        with torch.no_grad():
            embeddings = model(
                input_ids=tokenized_sequences, attention_mask=attention_mask
            )

        embeddings = embeddings[0].cpu().numpy()

        for seq_num, seq_len in zip_longest(range(len(embeddings)), seq_lens):
            # slice off first and last positions (special tokens)
            embedding = embeddings[seq_num][1 : seq_len + 1]
            assert (
                seq_len == embedding.shape[0]
            ), f"Sequence length mismatch: {seq_len} vs {embedding.shape[0]}"

            yield embedding

    @staticmethod
    def reduce_per_protein(embedding):
        return embedding.mean(axis=0)

    def embed(self, sequence: str) -> ndarray:
        [embedding] = self.embed_batch([sequence])
        return embedding



"""Visualize high dimensional data with t-SNE or UMAP projections"""

from bio_embeddings.project.tsne import tsne_reduce
from bio_embeddings.project.umap import umap_reduce

__all__ = [
    "tsne_reduce",
    "umap_reduce"
]


from umap import UMAP


def umap_reduce(embeddings, **kwargs):
    """Wrapper around :meth:`umap.UMAP` with defaults for bio_embeddings"""
    umap_params = dict()

    umap_params['n_components'] = kwargs.get('n_components', 3)
    umap_params['min_dist'] = kwargs.get('min_dist', .6)
    umap_params['spread'] = kwargs.get('spread', 1)
    umap_params['random_state'] = kwargs.get('random_state', 420)
    umap_params['n_neighbors'] = kwargs.get('n_neighbors', 15)
    umap_params['verbose'] = kwargs.get('verbose', 1)
    umap_params['metric'] = kwargs.get('metric', 'cosine')

    transformed_embeddings = UMAP(**umap_params).fit_transform(embeddings)

    return transformed_embeddings